import datetime as dt

from flask import Blueprint, request, current_app, jsonify, g
from flask_jwt_extended import create_access_token, create_refresh_token, jwt_refresh_token_required, get_jwt_identity, \
    jwt_optional

import dimensigon
from dimensigon import defaults
from dimensigon.domain.entities import Server, Catalog, User
from dimensigon.utils.helpers import get_now
from dimensigon.web import errors
from dimensigon.web.decorators import forward_or_dispatch, validate_schema, securizer, log_time
from dimensigon.web.helpers import check_param_in_uri
from dimensigon.web.json_schemas import login_post, healthcheck_post

blueprint_name = 'root'
root_bp = Blueprint(blueprint_name, __name__)


@root_bp.route('/')
def home():
    return {'message': 'Welcome to dimensigon'}


@root_bp.route('/healthcheck', methods=['GET', 'POST'])
@log_time('full')
@forward_or_dispatch()
@jwt_optional
@securizer
@validate_schema(POST=healthcheck_post)
@log_time('after validation')
def healthcheck():
    if request.method == 'POST' and isinstance(g.source, Server):
        try:
            heartbeat = dt.datetime.strptime(request.get_json()['heartbeat'], defaults.DATETIME_FORMAT)
        except:
            raise errors.InvalidDateFormat(request.get_json()['heartbeat'], defaults.DATETIME_FORMAT)
        exclude = request.get_json().get('exclude', [])
        current_app.cluster_manager.set_keepalive(g.source.id, heartbeat)

    catalog_ver = Catalog.max_catalog()
    data = {"version": dimensigon.__version__,
            "catalog_version": catalog_ver.strftime(defaults.DATEMARK_FORMAT) if catalog_ver else None,
            "scheduler": "running" if getattr(current_app.extensions.get('scheduler'), 'running',
                                              None) else "stopped",

            "services": [],

            }
    if not check_param_in_uri('human'):
        server = {'id': str(g.server.id), 'name': g.server.name}
        neighbours = [{'id': str(s.id), 'name': s.name} for s in Server.get_neighbours()]
        cluster = {'alive': [current_app.cluster_manager.cluster[i] for i in current_app.cluster_manager.cluster],
                   'in_coma': [current_app.cluster_manager.cluster[i] for i in current_app.cluster_manager.cluster if
                               i not in
                               current_app.cluster_manager.cluster.get_delta_keepalive(
                                   delta=current_app.dm.config.refresh_interval * defaults.COMA_NODE_FACTOR) and i != g.server.id]}
    else:
        server = g.server.name
        neighbours = sorted([s.name for s in Server.get_neighbours()])
        cluster = {'alive': sorted([getattr(Server.query.get(i), 'name', i) for i in current_app.cluster_manager.cluster]),
                   'in_coma': sorted([getattr(Server.query.get(i), 'name', i) for i in current_app.cluster_manager.cluster if
                               i not in
                               current_app.cluster_manager.cluster.get_delta_keepalive(
                                   delta=current_app.dm.config.refresh_interval * defaults.COMA_NODE_FACTOR) and i != g.server.id])}
    cluster.update(running=current_app.cluster_manager.running)
    data.update(server=server, neighbours=neighbours, cluster=cluster,
                now=get_now().strftime(defaults.DATETIME_FORMAT))



    return data


@root_bp.route('/ping', methods=['POST'])
@forward_or_dispatch()
def ping():
    req_data = request.json
    if req_data:
        req_data.update(dest_time=get_now().strftime(defaults.DATETIME_FORMAT))
        if 'servers' not in req_data:
            req_data.update(servers={})
    else:
        req_data = dict(dest_time=get_now().strftime(defaults.DATETIME_FORMAT))
    return req_data, 200


@root_bp.route('/login', methods=['POST'])
@forward_or_dispatch()
@validate_schema(login_post)
def login():
    user = User.get_by_user(user=request.json.get('username', None))
    password = request.json.get('password', None)
    try:
        if not user or not user.verify_password(password):
            return {"error": "Bad username or password"}, 401
    except TypeError:
        return {"error": "Bad username or password"}, 401

    # Use create_access_token() and create_refresh_token() to create our
    # access and refresh tokens
    ret = {
        'access_token': create_access_token(identity=str(user.id), fresh=True),
        'refresh_token': create_refresh_token(identity=str(user.id))
    }
    return jsonify(ret), 200


@root_bp.route('/refresh', methods=['POST'])
@forward_or_dispatch()
@jwt_refresh_token_required
def refresh():
    user = User.query.get(get_jwt_identity())
    ret = {
        'username': getattr(user, 'user', None),
        'access_token': create_access_token(identity=get_jwt_identity(), fresh=False)
    }
    return jsonify(ret), 200


@root_bp.route('/fresh-login', methods=['POST'])
@forward_or_dispatch()
@validate_schema(login_post)
def fresh_login():
    username = request.json.get('username', None)
    password = request.json.get('password', None)
    if username != 'test' or password != 'test':
        return jsonify({"msg": "Bad username or password"}), 401

    new_token = create_access_token(identity=username, fresh=True)
    ret = {'access_token': new_token}
    return jsonify(ret), 200


from unittest import TestCase, mock

from flask_jwt_extended import create_access_token

from dimensigon import defaults
from dimensigon.domain.entities import User
from dimensigon.domain.entities.base import SoftDeleteMixin, DistributedEntityMixin
from dimensigon.domain.entities.bootstrap import set_initial
from dimensigon.network.auth import HTTPBearerAuth
from dimensigon.web import create_app, db


class EntityWithSoftDelete(db.Model, SoftDeleteMixin):
    __tablename__ = 'E'
    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.Text)
    _old_name = db.Column(db.Text)


class DistEntityWithSoftDelete(db.Model, DistributedEntityMixin, SoftDeleteMixin):
    __tablename__ = 'DE'
    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.Text)
    _old_name = db.Column(db.Text)


    def to_json(self):
        data = super().to_json()
        data.update(id=self.id, name=self.name)
        return data


class Test(TestCase):
    def setUp(self):
        """Create and configure a new app instance for each test."""
        # create the app with common test config
        self.app = create_app('test')
        self.app_context = self.app.app_context()
        self.app_context.push()
        self.client = self.app.test_client()
        db.create_all()
        set_initial()
        self.auth = HTTPBearerAuth(create_access_token(User.get_by_user('root').id))

    def tearDown(self) -> None:
        db.session.remove()
        db.drop_all()
        self.app_context.pop()

    def test_create_entity_and_delete(self):
        e = EntityWithSoftDelete(id=1, name='first')

        self.assertFalse(e.deleted)

        db.session.add(e)
        db.session.commit()

        e.delete()

        self.assertTrue(e.deleted)

        self.assertEqual('first', e._old_name)
        self.assertNotEqual('first', e.name)

        db.session.commit()

        entities = EntityWithSoftDelete.query.all()

        self.assertEqual(0, len(entities))

    @mock.patch('dimensigon.domain.entities.base.random.choices')
    def test_create_distributed_entity_and_delete(self, mock_choices):

        mock_choices.return_value = list('abc')
        e = DistEntityWithSoftDelete(id=1, name='first', last_modified_at=defaults.INITIAL_DATEMARK)

        self.assertFalse(e.deleted)

        db.session.add(e)
        db.session.commit()

        dto = e.to_json()
        dto.pop('last_modified_at')
        self.assertDictEqual({'id': 1, 'name': 'first', 'deleted': False, '_old_name': None}, dto)

        e.delete()

        self.assertTrue(e.deleted)

        dto = e.to_json()
        dto.pop('last_modified_at')
        self.assertDictEqual({'id': 1, 'name': 'abc', 'deleted': True, '_old_name': 'first'}, dto)

        db.session.commit()

        dto = e.to_json(no_delete=True)
        dto.pop('last_modified_at')
        self.assertDictEqual({'id': 1, 'name': 'abc'}, dto)

        entities = EntityWithSoftDelete.query.all()

        self.assertEqual(0, len(entities))





import logging
import threading
import time
from contextlib import contextmanager

from sqlalchemy import event, inspect
from sqlalchemy.engine import Engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy.pool import Pool

from dimensigon.utils.helpers import get_distributed_entities, get_now
from dimensigon.web import db
from .action_template import ActionTemplate, ActionType
from .catalog import Catalog
from .dimension import Dimension
from .execution import StepExecution, OrchExecution
from .gate import Gate
from .locker import Locker, State, Scope
from .log import Log, Mode as LogMode
from .orchestration import Orchestration
from .parameter import Parameter
from .route import Route
from .schema_changes import SchemaChanges
from .server import Server
from .service import Service
from .software import Software, SoftwareServerAssociation
from .step import Step
from .transfer import Transfer, Status as TransferStatus
from .user import User

SCHEMA_VERSION = 5

_LOGGER = logging.getLogger('dimensigon.catalog')

__all__ = [
    "ActionTemplate",
    "ActionType",
    "Catalog",
    "Dimension",
    "StepExecution",
    "OrchExecution",
    "Gate",
    "Locker",
    "State",
    "Scope",
    "Log",
    "LogMode",
    "Orchestration",
    "Parameter",
    "Route",
    "SchemaChanges",
    "Service",
    "Server",
    "Software",
    "SoftwareServerAssociation",
    "Step",
    "Transfer",
    "TransferStatus",
    "User",
]

catalog = threading.local()


@contextmanager
def bypass_datamark_update(session=None):
    if session is None:
        session = db.session

    session.flush()
    update_datemark(False)
    try:
        yield None
    finally:
        session.flush()
        update_datemark(True)


def update_datemark(set):
    catalog.datemark = set


def set_events():
    for name, entity in get_distributed_entities():
        def receive_before_insert(mapper, connection, target):
            if not hasattr(catalog, 'data'):
                catalog.data = {}
            if getattr(catalog, 'datemark', True):
                target.last_modified_at = get_now()
            if not target.__class__ in catalog.data:
                catalog.data.update({target.__class__: target.last_modified_at})
            else:
                catalog.data.update(
                    {target.__class__: max(target.last_modified_at, catalog.data[target.__class__])})

        def receive_before_update(mapper, connection, target):
            # validate if target is updated
            it = inspect(target)
            columns = mapper.columns.keys()
            changed = any(
                [attr.history.has_changes() for attr in it.attrs if attr.key in columns])
            if changed:
                if not hasattr(catalog, 'data'):
                    catalog.data = {}
                if getattr(catalog, 'datemark', True):
                    target.last_modified_at = get_now()
                if not target.__class__ in catalog.data:
                    catalog.data.update({target.__class__: target.last_modified_at})
                else:
                    catalog.data.update(
                        {target.__class__: max(target.last_modified_at, catalog.data[target.__class__])})

        event.listen(entity, 'before_insert', receive_before_insert, propagate=False)
        event.listen(entity, 'before_update', receive_before_update, propagate=False)


set_events()


@event.listens_for(db.session, 'after_commit')
def receive_after_commit(session):
    # TODO: run this peace of code in a thread to allow the request end while executing the queries
    if hasattr(catalog, 'data'):
        # create a different session as the request session is already commited
        engine = db.get_engine()
        Session = sessionmaker(bind=engine)
        s = Session()
        for e, last_modified_at in catalog.data.items():
            # last_modified_at = s.query(func.max(e.last_modified_at)).scalar()
            c = s.query(Catalog).filter_by(entity=e.__name__).first()
            if c is None:
                c = Catalog(entity=e.__name__, last_modified_at=last_modified_at)
            else:
                if c.last_modified_at < last_modified_at:
                    c.last_modified_at = last_modified_at
            _LOGGER.debug(f'changed catalog {c}')
            s.add(c)
            s.commit()
            del c
        catalog.data = {}
        s.close()


@event.listens_for(Orchestration, 'refresh')
def receive_refresh(target, context, attrs):
    "listen for the 'refresh' event"
    target.init_on_load()


@event.listens_for(Pool, "connect")
def my_on_connect(dbapi_con, connection_record):
    # print("New DBAPI connection:", dbapi_con)
    dbapi_con.execute("PRAGMA journal_mode=WAL")
    # dbapi_con.execute("PRAGMA busy_timeout=10000")


_query_logger = logging.getLogger('dimensigon.query')


@event.listens_for(Engine, "before_cursor_execute")
def before_cursor_execute(conn, cursor, statement, parameters, context, executemany):
    conn.info.setdefault('query_start_time', []).append(time.time())


@event.listens_for(Engine, "after_cursor_execute")
def after_cursor_execute(conn, cursor, statement, parameters, context, executemany):
    total = time.time() - conn.info['query_start_time'].pop(-1)
    if total > 1:
        _query_logger.warning("Elapsed Time: %f\n%s\n%s", total, statement, parameters)



import inspect

import click
from prompt_toolkit import prompt


def resolve_callables(kwargs, params):
    prompt_kwargs = dict(kwargs)
    for kk, vv in kwargs.items():
        if callable(vv):
            sig = inspect.signature(vv)
            callable_kwargs = {kw: params.get(kw, None) for kw in sig.parameters.keys()}
            prompt_kwargs[kk] = vv(**callable_kwargs)
    return prompt_kwargs


def set_parameter(parameter, text, entity, load=None):
    if text == '':
        text = None
    elif text == "''" or text == '""':
        text = ""
    else:
        if callable(load):
            text = load(text)
    if text != entity.get(parameter, None):
        entity[parameter] = text
        return True
    else:
        return False


def prompt_parameter(parameter, entity, form, parent_prompt) -> bool:
    prompt_kwargs = resolve_callables(form[parameter], entity)

    edit = prompt_kwargs.pop('edit', False)
    value = entity.get(parameter, '')
    if 'validator' in prompt_kwargs and hasattr(prompt_kwargs['validator'], 'dump'):
        value = prompt_kwargs['validator'].dump(value)
    else:
        value = str(value) if value is not None else ''

    if edit:
        text = click.edit(str(value) if value != '' else '')
    else:
        text = prompt(
            f"{parent_prompt}.{parameter}>{'>' if prompt_kwargs.get('multiline', False) else ''} ",
            default=value,
            **prompt_kwargs)
    return set_parameter(parameter, text, entity,
                         prompt_kwargs['validator'].load if 'validator' in prompt_kwargs and hasattr(
                             prompt_kwargs['validator'], 'load') else None)




import ast
import json

from prompt_toolkit.validation import Validator, ValidationError


class JSONValidator(Validator):

    @staticmethod
    def load(text):
        return json.loads(text)

    @staticmethod
    def dump(data):
        return json.dumps(data)

    def validate(self, document):
        text = document.text

        if text:
            try:
                self.load(text)
            except Exception as e:
                raise ValidationError(message=str(e))


class BoolValidator(Validator):
    true = ['true', 'yes', 'y']
    false = ['false', 'no', 'n']

    @classmethod
    def load(cls, x: str):
        if x.strip().lower() in cls.true:
            return True
        elif x.strip().lower() in cls.false:
            return False
        else:
            raise ValueError

    @classmethod
    def dump(cls, data):
        if data:
            return cls.true[0]
        else:
            return cls.false[0]

    def validate(self, document):
        text = document.text

        if text:
            try:
                self.load(text)
            except:
                raise ValidationError(message=f'Not a valid bool value ({",".join(self.true)}|{",".join(self.false)})')


class IntValidator(Validator):

    @staticmethod
    def load(x: str):
        return int(x) if x is not None else x

    @staticmethod
    def dump(data):
        return str(data)

    def validate(self, document):
        text = document.text

        if text:
            try:
                self.load(text)
            except:
                raise ValidationError(message='Not a valid integer')


class ChoiceValidator(Validator):

    def __init__(self, choices) -> None:
        super().__init__()
        self.choices = choices

    def validate(self, document):
        text = document.text

        if text and text not in self.choices:
            raise ValidationError(message='Invalid choice. Choose from ' + ', '.join(self.choices),
                                  cursor_position=0)


class ListConverter(Validator):

    def __init__(self, separator=None) -> None:
        self.separator = separator

    @staticmethod
    def dump(data):
        return ' '.join(data or [])

    def load(self, x: str):
        return x.split(sep=self.separator)

    def validate(self, document):
        return document.text


class ListValidator(Validator):


    @staticmethod
    def load(x: str):
        v = ast.literal_eval(x) if x is not None else '[]'
        if v and not isinstance(v, list):
            raise ValueError()
        return v

    def validate(self, document):
        text = document.text

        if text:
            try:
                self.load(text)
            except Exception as e:
                raise ValidationError(message='Not a valid list. ' + str(e))



import datetime
import json
import logging
import typing as t

from pkg_resources import parse_version

import dimensigon.use_cases.routing as routing
import dimensigon.web.network as ntwrk
from dimensigon import __version__ as dm_version, defaults
from dimensigon.domain.entities import Server, Catalog
from dimensigon.use_cases.helpers import get_root_auth
from dimensigon.use_cases.use_cases import run_elevator, get_software, upgrade_catalog_from_server
from dimensigon.utils import asyncio
from dimensigon.utils.asyncio import create_task
from dimensigon.utils.helpers import get_now
from dimensigon.web import db, errors
from dimensigon.web.decorators import run_as

logger = logging.getLogger('dimensigon.background')
catalog_logger = logging.getLogger('dimensigon.catalog')
upgrader_logger = logging.getLogger('dimensigon.upgrader')


#
# @run_as('root')
# def process_get_new_version_from_gogs(app=None):
#     """
#     checks if new version in repo
#
#     Parameters
#     ----------
#     app:
#         app to load the context
#     timeout_wait_transfer:
#         timeout waiting tranfer file to end.
#     refresh_interval:
#         time period to check if tranfer ended. Normally, used for test purposes
#
#     Returns
#     -------
#
#     """
#     upgrader_logger.info('Starting Upgrade Process')
#     base_url = os.environ.get('GIT_REPO') \
#                or current_app.config.get('GIT_REPO') \
#                or 'https://ca355c55-0ab0-4882-93fa-331bcc4d45bd.pub.cloud.scaleway.com:3000'
#     releases_uri = '/dimensigon/dimensigon/releases'
#     try:
#         r = requests.get(base_url + releases_uri, verify=current_app.config['SSL_VERIFY'], timeout=10)
#     except (requests.exceptions.Timeout, requests.exceptions.ConnectionError) as e:
#         r = None
#         upgrader_logger.info('Unable to contact to main repo')
#
#     # get new versions from repo
#     if r and r.status_code == 200:
#         # get current software
#         gogs_versions = {}
#
#         html_content = r.text
#         soup = BeautifulSoup(html_content, 'html.parser')
#         for li in soup.find(id='release-list').find_all('li'):
#             version = li.h4.a.get_text(strip=True)
#             uris = [a.attrs['href'] for a in li.find('div', class_='download').find_all('a') if
#                     a.attrs['href'].endswith('tar.gz')]
#             if len(uris) > 0:
#                 gogs_versions.update({parse_version(version): uris[0]})
#         current_version = parse_version(dm_version)
#         new_versions = [gogs_ver for gogs_ver in gogs_versions if gogs_ver > current_version]
#
#         if new_versions:
#             new_version = max(new_versions)
#             upgrader_logger.info(f"Downloading version {new_version} from outside world")
#
#             r = requests.get(base_url + gogs_versions[new_version],
#                              verify=current_app.config['SSL_VERIFY'])
#             filename = get_filename_from_cd(
#                 r.headers.get(
#                     'content-disposition')) or f"dimensigon-{gogs_versions[new_version].rsplit('/', 1)[-1]}"
#             os.makedirs(os.path.join(current_app.config['SOFTWARE_REPO'], 'dimensigon'), exist_ok=True)
#             file = os.path.join(current_app.config['SOFTWARE_REPO'], 'dimensigon', filename)
#             try:
#                 open(file, 'wb').write(r.content)
#             except Exception as e:
#                 upgrader_logger.exception(f"Unable to save {file}")
#             else:
#                 run_elevator(file, new_version, upgrader_logger)
#     else:
#         upgrader_logger.debug(f"No version to upgrade")


async def _async_get_neighbour_catalog_data_mark(cluster_heartbeat_id: str = None) -> t.Dict[Server, ntwrk.Response]:
    server_responses = {}
    servers = Server.get_neighbours()
    catalog_logger.debug(f"Neighbour servers to check: {[s.name for s in servers]}")
    auth = get_root_auth()

    if cluster_heartbeat_id is None:
        cluster_heartbeat_id = get_now().strftime(defaults.DATETIME_FORMAT)
    for server in servers:
        server_responses[server] = create_task(ntwrk.async_post(server, 'root.healthcheck', auth=auth,
                                                                json={'exclude': [s.id for s in servers],
                                                                      'heartbeat': cluster_heartbeat_id}))

    for server, future in server_responses.items():
        server_responses[server] = await future
        if server_responses[server].ok:
            id_response = server_responses[server].msg.get('server', {}).get('id', '')
            if id_response and str(server.id) != id_response:
                server_responses[server].exception = errors.HealthCheckMismatch(
                    expected={'id': str(server.id), 'name': server.name},
                    actual=server_responses[server].msg.get('server', {}))
                server_responses[server].msg = None
                server_responses[server].code = None

        catalog_logger.log(1,
            f"Response from server {server.name}: {json.dumps(server_responses[server].to_dict(), indent=4)}")

    return server_responses


def upgrade_version(data: t.Dict[Server, ntwrk.Response]):
    mayor_version, mayor_server = None, None
    for server, response in data.items():
        if response.code == 200 and 'version' in response.msg:
            remote_version = parse_version(response.msg['version'])
            if remote_version > parse_version(dm_version):
                if mayor_version is None or mayor_version < remote_version:
                    mayor_version, mayor_server = remote_version, server
    if mayor_version:
        catalog_logger.info(f'Found mayor version on server {mayor_server}. Upgrading version first')
        file, v = get_software(mayor_server, get_root_auth())
        if file:
            run_elevator(file, mayor_version, catalog_logger)
            return True
    return False


def update_catalog(data: t.Dict[Server, ntwrk.Response]):
    reference_server = None
    catalog_ver = db.session.query(db.func.max(Catalog.last_modified_at)).scalar()
    if catalog_ver:
        for server, response in data.items():
            if response.code == 200 and 'catalog_version' in response.msg:
                new_catalog_ver = datetime.datetime.strptime(response.msg['catalog_version'],
                                                             defaults.DATEMARK_FORMAT)
                if new_catalog_ver > catalog_ver:
                    if response.msg['version'] == dm_version:
                        reference_server = server
                        catalog_ver = new_catalog_ver
                    else:
                        catalog_logger.debug(
                            f"Server {server} has different software version {response.msg['version']}")
            else:
                msg = f"Error while trying to get healthcheck from server {server.name}. "
                if response.code:
                    msg = msg + f"Response from server (code {response.code}): {response.msg}"
                else:
                    msg = msg + f"Exception: {response.code}"
                catalog_logger.warning(msg)
        if reference_server:
            catalog_logger.info(f"New catalog found from server {reference_server.name}: {catalog_ver}")
            upgrade_catalog_from_server(reference_server)
        else:
            catalog_logger.debug(f"No server with higher catalog found")


@run_as('root')  # takes the app argument and pushes the context
def process_catalog_route_table(app=None, upgrade_catalog=True):
    # app will be used for the run_as decorator
    from dimensigon.web.api_1_0.urls.use_cases import delete_old_temp_servers
    delete_old_temp_servers()
    asyncio.run(routing.async_update_routes_send(discover_new_neighbours=True, check_current_neighbours=True,
                                                 max_num_discovery=3))

    catalog_logger.debug("Starting check catalog from neighbours")

    # cluster information
    cluster_hearthbeat_id = get_now().strftime(defaults.DATETIME_FORMAT)
    # check version upgrade before catalog upgrade to match database revision
    if upgrade_catalog:
        data = asyncio.run(_async_get_neighbour_catalog_data_mark(cluster_hearthbeat_id))
        if not upgrade_version(data):
            update_catalog(data)
    db.session.commit()



import argparse
import base64
import datetime as dt
import functools
import logging
import os
import platform
import random
import sys
import time
from dataclasses import dataclass

import coolname
import prompt_toolkit
import requests
import rsa
import yaml
from flask import Flask
from flask_jwt_extended import create_access_token

from dimensigon import defaults
from dimensigon.core import Dimensigon
from dimensigon.use_cases.helpers import get_root_auth

basedir = os.path.abspath(os.path.dirname(__file__))

PLATFORM = platform.system()


from dimensigon.domain.entities import *
from dimensigon.web.network import pack_msg2, unpack_msg2
from dimensigon.web import create_app, db

from dimensigon.use_cases.use_cases import upgrade_catalog
from dimensigon.utils.helpers import generate_symmetric_key, generate_dimension, get_now

app: Flask = create_app(os.getenv('FLASK_CONFIG') or 'default')


def new(dm: Dimensigon, name: str):
    dm.create_flask_instance()
    with dm.flask_app.app_context():
        from cryptography import x509
        from cryptography.hazmat.backends import default_backend
        from cryptography.hazmat.primitives import hashes, serialization
        from cryptography.x509.oid import NameOID
        import datetime
        Server.set_initial()
        count = Dimension.query.count()

        if count > 0:
            exit("Only one dimension can be created")

        dim_name = name or coolname.generate_slug(2)
        dim = generate_dimension(dim_name)

        private_key = serialization.load_pem_private_key(dim.private.save_pkcs1(), password=None,
                                                         backend=default_backend())
        dim.current = count == 0
        db.session.add(dim)

        now = get_now()

        subject = issuer = x509.Name([
            x509.NameAttribute(NameOID.COUNTRY_NAME, u"LU"),
            x509.NameAttribute(NameOID.STATE_OR_PROVINCE_NAME, u"California"),
            x509.NameAttribute(NameOID.LOCALITY_NAME, u"San Francisco"),
            x509.NameAttribute(NameOID.COMMON_NAME, dim_name),
            x509.NameAttribute(NameOID.ORGANIZATION_NAME, u"KnowTrade S.L."),

        ])

        cert = x509.CertificateBuilder().subject_name(subject) \
            .issuer_name(issuer) \
            .not_valid_before(now) \
            .not_valid_after(now + datetime.timedelta(days=365 * 10)) \
            .serial_number(x509.random_serial_number()) \
            .public_key(private_key.public_key()) \
            .add_extension(x509.BasicConstraints(ca=True, path_length=None), critical=False, ) \
            .sign(private_key=private_key, algorithm=hashes.SHA256(),
                  backend=default_backend()
                  )

        ssl_dir = os.path.join(dm.config.config_dir, defaults.SSL_DIR)
        os.makedirs(ssl_dir, exist_ok=True)

        with open(os.path.join(ssl_dir, defaults.CERT_FILE), "wb") as f:
            f.write(cert.public_bytes(serialization.Encoding.PEM))

        with open(os.path.join(ssl_dir, defaults.KEY_FILE), 'wb') as file:
            file.write(private_key.private_bytes(
                encoding=serialization.Encoding.PEM,
                format=serialization.PrivateFormat.TraditionalOpenSSL,
                encryption_algorithm=serialization.NoEncryption(),

            ))

        os.chmod(os.path.join(ssl_dir, defaults.KEY_FILE), 0o600)

        db.session.commit()

        user = User.get_by_user('root')
        if user is None:
            User.set_initial()
            user = User.get_by_user('root')

        p = False
        p2 = True
        while p != p2:
            p = prompt_toolkit.prompt("Password for root user: ", is_password=True)
            p2 = prompt_toolkit.prompt("Re-type same password: ", is_password=True)
            if p != p2:
                print('Password mismatch')
        user.set_password(p)
        del p, p2

        db.session.commit()

        print(f"New dimension created successfully")
        print("")
        print("----- JOIN TOKEN (valid for {} minutes) -----".format(defaults.JOIN_TOKEN_EXPIRE_TIME))
        token(dm, dim.id)
        print("---------------- END TOKEN --------------------")


def server_port():
    if 'PORT' not in os.environ:
        return '5000'
    else:
        return os.environ['PORT']


def server_bind_address():
    if 'HTTP_HOST' not in os.environ:
        return '127.0.0.1'
    else:
        return os.environ['HTTP_HOST']


def app_context(f):
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        with app.app_context():
            f(*args, **kwargs)

    return wrapper


def join(dm: Dimensigon, server: str, token: str, port: int = None, ssl: bool = True, verify: bool = False):
    logger = logging.getLogger('dimensigon.join')
    dm.create_flask_instance()
    with dm.flask_app.app_context():
        Server.set_initial()
        db.session.commit()
        protocol = "https" if ssl else "http"

        resp = None
        times = 5
        for i in range(times):
            try:
                resp = requests.get(f"{protocol}://{server}:{port}/api/v1.0/join/public",
                                    headers={'Authorization': 'Bearer ' + token},
                                    verify=verify, timeout=20)
            except requests.exceptions.ConnectionError as e:
                logger.error(f"Unable to contact to {server}.")
            except requests.exceptions.Timeout as e:
                logger.error(f"Timeout of 20s reached while trying to contact to {server}.")
            except Exception as e:
                logger.exception(f"Error trying to get dimensigon public key.")
            else:
                if resp.status_code != 200:
                    if resp.status_code in (401, 422):
                        logger.error(f"Error trying to get dimensigon public key: {resp.status_code}, {resp.content}")
                        break
                    else:
                        logger.error(f"Error trying to get dimensigon public key: {resp.status_code}, {resp.content}")
                else:
                    break
            if i < times - 1:
                d = int(random.random() * 25 * (i + 1))
                logger.info(f"Retrying in {d} seconds.")
                time.sleep(d)
            else:
                sys.exit(1)

        pub_key = rsa.PublicKey.load_pkcs1(resp.content)

        # Generate Public and Private Temporal Keys
        tmp_pub, tmp_priv = rsa.newkeys(2048, poolsize=2)
        symmetric_key = generate_symmetric_key()
        s = Server.get_current()
        data = s.to_json(add_gates=True)
        data = pack_msg2(data=data, pub_key=pub_key, priv_key=tmp_priv, symmetric_key=symmetric_key, add_key=True)
        data.update(my_pub_key=tmp_pub.save_pkcs1().decode('ascii'))
        logger.info("Joining to dimension...")

        resp = None
        times = 5
        for i in range(times):
            try:
                resp = requests.post(f"{protocol}://{server}:{port}/api/v1.0/join", json=data,
                                     headers={'Authorization': 'Bearer ' + token}, verify=verify, timeout=45)

                from dimensigon.web import errors
                if resp.ok:
                    break
                else:
                    logger.error(f"Error while trying to join. Error: {resp.status_code}, "
                                 f"{resp.content}")
            except Exception as e:
                logger.exception(f"Error while trying to join.")
                resp = None
            if i < times - 1:
                d = int(random.random() * 25 * (i + 1))
                logger.info(f"Retrying in {d} seconds.")
                time.sleep(d)

        if resp is None:
            resp_data = {}
        elif resp.status_code != 200:
            db.session.rollback()
            logger.info(f"Error while trying to join the dimension: {resp.status_code}, {resp.content}")
            resp_data = {}
        else:
            resp_data = unpack_msg2(resp.json(), pub_key=pub_key, priv_key=tmp_priv,
                                    symmetric_key=symmetric_key)
            logger.log(1, resp_data)
        if 'Dimension' in resp_data:
            dim = Dimension.from_json(resp_data.pop('Dimension'))
            dim.current = True
            db.session.add(dim)

            keyfile_content = base64.b64decode(resp_data.pop('keyfile').encode())
            with open(dm.config.http_conf['keyfile'], 'wb') as fh:
                fh.write(keyfile_content)
                del keyfile_content
            certfile_content = base64.b64decode(resp_data.pop('certfile').encode())
            with open(dm.config.http_conf['certfile'], 'wb') as fh:
                fh.write(certfile_content)
                del certfile_content

            logger.info('Updating Catalog...')
            reference_server_id = resp_data.pop('me')
            # remove catalog
            for c in Catalog.query.all():
                db.session.delete(c)
            try:
                upgrade_catalog(resp_data['catalog']) # implicit commit
            except Exception as e:
                logger.exception(f"Unable to upgrade catalog.")
                exit(1)
            else:
                logger.info('Catalog updated.')
            # set reference server as a neighbour
            reference_server = Server.query.get(reference_server_id)
            if not reference_server:
                db.session.rollback()
                logger.info(f"Server id {reference_server_id} not found in catalog.")
                exit(1)
            Route(destination=reference_server, cost=0)
            # update_route_table_cost(True)
            Parameter.set("join_server", f'{reference_server.id}')

            resp = None
            times = 5
            for i in range(times):
                try:
                    resp = requests.post(f"{protocol}://{server}:{port}/api/v1.0/join/acknowledge/{s.id}",
                                         headers={'Authorization': 'Bearer ' + token}, verify=verify, timeout=120)

                    from dimensigon.web import errors
                    if resp.ok:
                        break
                    elif resp.status_code == 404:
                        if resp.json().get('error', {}).get('type', None) == 'EntityNotFound':
                            break
                    else:
                        logger.error(f"Error while trying to join. Error: {resp.status_code}, "
                                     f"{resp.content}")
                except Exception as e:
                    logger.exception(f"Error while trying to join.")
                    resp = None
                if i < times - 1:
                    d = int(random.random() * 25 * (i + 1))
                    logger.info(f"Retrying in {d} seconds.")
                    time.sleep(d)

            if resp.ok:
                logger.info('Joined to the dimension.')
            else:
                logger.error("Unable to confirm join.")

            db.session.commit()
        else:
            logger.error(f"No dimension in response data.")


def token(dm: Dimensigon, dimension_id_or_name: str, applicant=None, expire_time=None):
    dm.create_flask_instance()
    with dm.flask_app.app_context():
        if dimension_id_or_name is not None:
            dim = Dimension.query.get(dimension_id_or_name)
            if dim is None:
                dim = Dimension.query.filter_by(name=dimension_id_or_name)
                if dim is None:
                    exit(f"{dimension_id_or_name} is not a valid dimension")
        else:
            count = Dimension.query.count()
            if count == 1:
                dim = Dimension.query.all()[0]
            else:
                exit("No dimension specified. Please specify a dimension")

        if dimension_id_or_name:
            dm.flask_app.config['SECRET_KEY'] = dim.id
        else:
            if dm.flask_app.config['SECRET_KEY'] != dim.id:
                exit('Secret key mismatch')

        join_user = User.get_by_user('join')
        if not join_user:
            exit("Unable to create join token. Populate database first.")
        print(create_access_token(join_user.id,
                                  expires_delta=dt.timedelta(
                                      minutes=expire_time or defaults.JOIN_TOKEN_EXPIRE_TIME),
                                  user_claims={'applicant': applicant}))


def catalog(dm: Dimensigon, ip, port, http=False):
    import dimensigon.dshell.network as dshell_ntwrk
    dm.create_flask_instance()
    with dm.flask_app.app_context():
        print("Updating catalog...")
        catalog_datamark = Catalog.max_catalog(str)

        resp = dshell_ntwrk.request('get', dshell_ntwrk.generate_url('api_1_0.catalog',
                                                                     view_data=dict(data_mark=catalog_datamark),
                                                                     ip=ip,
                                                                     port=port,
                                                                     scheme='http' if http else 'https'),
                                    auth=get_root_auth())
        if resp.ok:

            try:
                upgrade_catalog(resp.msg)
            except Exception as e:
                exit(f"Unable to upgrade data. Exception: {e}")
        else:
            exit(f"Unable to get catalog from {resp.url}: {resp}")


def run(dm: Dimensigon):
    # check if there is a dimension
    result = dm.engine.execute(Dimension.__table__.select())
    count = len(result.fetchall())
    if count == 0:
        exit("No dimension created. Create or join to a dimension")
    dm.start()


def get_arguments() -> argparse.Namespace:
    from dimensigon import __version__
    parser = argparse.ArgumentParser(prog='dimensigon')
    parser.add_argument('--version', action='version',
                        version='%(prog)s {version}'.format(version=__version__))
    parser.add_argument(
        "-c",
        "--config-dir",
        metavar="path_to_config_dir",
        help="Directory that contains the dimensigon configuration",
    )
    parser.add_argument(
        "--pid-file",
        metavar="path_to_pid_file",
        default=None,
        help="Path to PID file useful for running as daemon. If not set, default config path will be used",
    )
    parser.add_argument(
        "--port",
        metavar="listen_port",
        default=defaults.DEFAULT_PORT,
        help="Listen HTTP port",
    )
    parser.add_argument(
        "--ip",
        default=None,
        dest='ips',
        action='append',
        help="IP to listen",
    )
    parser.add_argument(
        "--key-file",
        dest='keyfile',
        default=None,
        help="SSL key file",
    )
    parser.add_argument(
        "--cert-file",
        dest='certfile',
        default=None,
        help="SSL certificate file",
    )
    parser.add_argument(
        "--threads",
        default=None,
        help="The number of worker threads for handling requests.",
    )
    parser.add_argument(
        "--debug",
        action='store_true',
        help="Runs the Flask app without gunicorn.",
    )
    parser.add_argument(
        "--access-logfile",
        dest='accesslog',
        metavar="FILE",
        help="The Access log file to write to.",
    )
    parser.add_argument(
        "--error-logfile",
        dest='errorlog',
        metavar="FILE",
        help="The Error log file to write to.",
    )
    parser.add_argument(
        "--logconfig-file",
        metavar="FILE",
        type=argparse.FileType('r'),
        help="The log configuration in yaml format to load.",
    )
    parser.add_argument(
        "--dm-refresh-interval",
        help="Period time in minutes to execute the routing and catalog refresh.",
        default=defaults.REFRESH_PERIOD,
        type=int
    )
    parser.add_argument(
        "--force-scan",
        help="forces the process to check current and scan for new neighbours.",
        action='store_true',
    )
    parser.add_argument(
        "--flask",
        action='store_true',
        help="run the process with flask http.",
    )
    if os.name == "posix":
        parser.add_argument(
            "--daemon", action="store_true", help="Run Dimensigon as daemon"
        )

    subparser = parser.add_subparsers(dest='command')
    join_parser = subparser.add_parser("join", help="Joins to the dimension.")
    join_parser.add_argument(
        "SERVER",
        help="Reference Server which will allow to join the dimension. You must specify server name or IP"
    )
    join_parser.add_argument(
        "TOKEN",
        help="Authentication Token used for authentication to the reference server"
    )
    join_parser.add_argument(
        '--port', '-p',
        help="port used to contact the server. Defaults to 5000",
        default=defaults.DEFAULT_PORT,
    )
    join_parser.add_argument(
        '--no-ssl',
        dest='ssl',
        help="makes connection with HTTP protocol", action='store_false'
    )
    join_parser.add_argument(
        '--verify',
        help="verifies certificate",
        action='store_true'
    )


    new_parser = subparser.add_parser('new', help="Creates a new dimension")
    new_parser.add_argument('dimension', nargs='?', default=None, help="name of the dimension")

    token_parser = subparser.add_parser('token', help="Generates a join token.")
    token_parser.add_argument(
        'dimension',
        nargs='?',
        default=None
    )
    token_parser.add_argument("--expire-time",
                              metavar='MINUTES',
                              type=int,
                              help="Join token expire time in minutes")
    token_parser.add_argument(
        '--applicant',
        help="applicant identifier used for join operations",
        action='store'
    )

    catalog_parser = subparser.add_parser("catalog", help="Forces updates catalog with specified ip and port.")
    catalog_parser.add_argument("IP",
                                help="ip to force catalog update")
    catalog_parser.add_argument("--port",
                                type=int,
                                default=defaults.DEFAULT_PORT)
    catalog_parser.add_argument("--http",
                                action='store_true')

    arguments = parser.parse_args()

    if os.name != "posix":
        setattr(arguments, "daemon", False)

    return arguments


@dataclass
class RuntimeConfig:
    config_dir: str = None
    debug: bool = None
    pid_file: str = None
    port: str = None
    ips: list = None
    keyfile: str = None
    certfile: str = None
    threads: int = None
    daemon: bool = None
    accesslog: str = None
    errorlog: str = None
    logconfig: dict = None
    flask: bool = None
    refresh_interval: int = None
    force_scan: bool = None


def main():
    args = get_arguments()

    from dimensigon import bootstrap

    dm = bootstrap.setup_dm(RuntimeConfig(config_dir=args.config_dir,
                                          debug=args.debug,
                                          pid_file=args.pid_file,
                                          port=args.port,
                                          ips=args.ips,
                                          keyfile=args.keyfile,
                                          certfile=args.certfile,
                                          threads=args.threads,
                                          daemon=args.daemon,
                                          accesslog=args.accesslog,
                                          errorlog=args.errorlog,
                                          logconfig=yaml.load(args.logconfig_file,
                                                              Loader=yaml.FullLoader) if args.logconfig_file else {},
                                          flask=args.flask,
                                          refresh_interval=args.dm_refresh_interval,
                                          force_scan=args.force_scan))

    if args.command is None:
        run(dm)
    elif args.command == 'join':
        join(dm, server=args.SERVER, token=args.TOKEN, port=args.port, ssl=args.ssl, verify=args.verify)
    elif args.command == 'token':
        token(dm, dimension_id_or_name=args.dimension, applicant=args.applicant, expire_time=args.expire_time)
    elif args.command == 'new':
        new(dm, args.dimension)
    elif args.command == 'catalog':
        catalog(dm, ip=args.IP, port=args.port, http=args.http)
    else:
        exit("Use -h to show help")


if __name__ == '__main__':
    main()



from flask import request, g, current_app
from flask_jwt_extended import jwt_required
from flask_restful import Resource

from dimensigon.domain.entities import Server
from dimensigon.use_cases import routing
from dimensigon.web import errors, db
from dimensigon.web.decorators import securizer, forward_or_dispatch, lock_catalog, validate_schema
from dimensigon.web.helpers import filter_query, check_param_in_uri
from dimensigon.web.json_schemas import server_patch, servers_delete


class ServerList(Resource):

    @forward_or_dispatch()
    @jwt_required
    @securizer
    def get(self):
        query = filter_query(Server, request.args)
        return [s.to_json(add_gates=check_param_in_uri('gates'),
                          human=check_param_in_uri('human'),
                          no_delete=True,
                          add_ignore=True) for s in
                query.all()]

    @forward_or_dispatch()
    @jwt_required
    @securizer
    @validate_schema(servers_delete)
    @lock_catalog
    def delete(self):
        servers = [Server.query.get_or_raise(s_id) for s_id in request.get_json()['server_ids']]
        acquired = routing._lock.acquire(timeout=15)
        if acquired:
            current_app.logger.debug(f"Routing Lock acquired for deletion of servers {servers}")
        else:
            current_app.logger.debug(f"Unable to lock Routing Lock. Force deletion of servers {servers}")
        try:
            for server in servers:
                if server == g.server:
                    raise errors.ServerDeleteError
                # remove associated routes
                db.session.delete(server.route)
                server.delete()
            db.session.commit()
        finally:
            if acquired:
                routing._lock.release()

        return {}, 204


class ServerResource(Resource):
    @forward_or_dispatch()
    @jwt_required
    @securizer
    def get(self, server_id):
        return Server.query.get_or_raise(server_id).to_json(add_gates=check_param_in_uri('gates'),
                                                          human=check_param_in_uri('human'),
                                                          no_delete=True,
                                                          add_ignore=True)

    @forward_or_dispatch()
    @jwt_required
    @securizer
    @validate_schema(server_patch)
    @lock_catalog
    def patch(self, server_id):
        json_data = request.get_json()

        server = Server.query.get_or_raise(server_id)
        new_granules = json_data.get('granules', [])
        if 'all' in new_granules:
            raise errors.KeywordReserved("'all' is a reserved granule")

        server.granules = list(set(server.granules) | set(new_granules))

        for gate in json_data.get('gates'):
            server.add_new_gate(gate['dns_or_ip'], gate['port'], gate.get('hidden'))

        if 'ignore_on_lock' in json_data:
            server.l_ignore_on_lock = json_data.get('ignore_on_lock')

        db.session.commit()

        return {}, 204

    @forward_or_dispatch()
    @jwt_required
    @securizer
    @lock_catalog
    def delete(self, server_id):
        server = Server.query.get_or_raise(server_id)
        acquired = routing._lock.acquire(timeout=15)
        if acquired:
            current_app.logger.debug(f"Routing Lock acquired for deletion of server {server}")
        else:
            current_app.logger.debug(f"Unable to lock Routing Lock. Force deletion of server {server}")
        try:
            if server == g.server:
                raise errors.ServerDeleteError
            # remove associated routes
            db.session.delete(server.route)
            server.delete()
            db.session.commit()
        except:
            raise
        finally:
            if acquired:
                routing._lock.release()

        return {}, 204




import configparser
import logging
import os

from dimensigon import defaults
from dimensigon.dshell import environ as env
from dimensigon.dshell import network as ntwrk

_logger = logging.getLogger('dshell')

def load_config_file(file=None):
    data = dict(USERNAME=None, TOKEN=None, SERVER=None, PORT=None)
    config = configparser.ConfigParser()
    if file and os.path.exists(os.path.expanduser(file)):
        config.read(os.path.expanduser(file))
        data['USERNAME'] = config.get('AUTH', 'username', fallback=None)
        data['TOKEN'] = config.get('AUTH', 'token', fallback=None)
        data['SERVER'] = config.get('REMOTE', 'server', fallback=None)
        data['PORT'] = config.getint('REMOTE', 'port', fallback=None)
    return data


def save_config_file(file=None, username=None, token=None, server=None, port=None):
    config = configparser.ConfigParser()
    if file:
        config.read(file)
        if not config.has_section('AUTH'):
            config.add_section('AUTH')
        if not config.has_section('REMOTE'):
            config.add_section('REMOTE')
        if username:
            config.set('AUTH', 'username', username)
        if token:
            config.set('AUTH', 'token', token)
        if server:
            config.set('REMOTE', 'server', server)
        if port:
            config.set('REMOTE', 'port', str(port))
        with open(file, 'w') as fd:
            config.write(fd)
    _logger.info(f"saved data into {file}")


def bootstrap_environ(run_config):
    config_file = run_config['--config-file']
    data = load_config_file(os.path.expanduser(config_file))

    server = run_config['--server'] or os.environ.get('DM_SERVER') or data.get('SERVER', None) or '127.0.0.1'
    port = run_config['--port'] or os.environ.get('DM_PORT') or data.get('PORT', None) or defaults.DEFAULT_PORT

    env.set_dict_in_environ({'SERVER': server, 'PORT': port, 'SCHEME': 'https', 'SSL_VERIFY': False,
                             'FILE_HISTORY': '~/.dshell_history', 'DEBUG': False, 'CONFIG_FILE': config_file})

    # auth data
    username = run_config['--username'] or os.environ.get('DM_USERNAME') or data.get('USERNAME', None)
    password = run_config['--password'] or os.environ.get('DM_PASSWORD')
    refresh_token = run_config['--token'] or os.environ.get('DM_TOKEN') or data.get('TOKEN')

    if refresh_token:
        env._refresh_token = refresh_token
    if username:
        env._username = username
        if password:
            ntwrk.login(username, password)

    env.set('DEBUG', run_config.get('--debug', False))



import datetime as dt
import functools
import re
from functools import partial
from unittest import TestCase
from unittest.mock import patch

import responses
from aioresponses import aioresponses, CallbackResult

from dimensigon.__main__ import join, Gate, token as generate_token, Locker, User
from dimensigon.domain.entities import Server, Dimension, Catalog
from dimensigon.utils.helpers import generate_dimension
from dimensigon.web import create_app, db


class TestApi(TestCase):

    @patch('dimensigon.domain.entities.get_now')
    def setUp(self, mock_now):
        """Create and configure a new app instance for each test."""
        mock_now.return_value = dt.datetime(2019, 4, 1, tzinfo=dt.timezone.utc)
        # create the app with common test config
        self.app_join = create_app('test')
        self.app_join.config['SECURIZER'] = True
        self.app_join.config['SERVER_NAME'] = 'new'
        with self.app_join.app_context():
            db.create_all()
            Locker.set_initial()
            User.set_initial()
            Server.set_initial()
            me = Server.get_current()
            # remove Gates
            [db.session.delete(g) for g in me.gates]
            g = Gate(me, port=8000, dns='new')
            db.session.add(g)
            db.session.commit()
            self.server_new = Server.get_current().to_json()

        self.app = create_app('test')
        self.app.config['SERVER_NAME'] = 'node1'
        self.app.config['SECURIZER'] = True
        mock_now.return_value = dt.datetime(2019, 3, 1, tzinfo=dt.timezone.utc)
        with self.app.app_context():
            db.create_all()
            Locker.set_initial()
            User.set_initial()
            Server.set_initial()
            me = Server.get_current()
            # remove Gates
            [db.session.delete(g) for g in me.gates]
            g = Gate(me, port=8000, dns='node1')
            db.session.add(g)
            db.session.commit()
            dim = generate_dimension('test')
            dim.current = True
            db.session.add(dim)
            db.session.commit()
            self.dimension = dim.to_json()
            self.server_node1 = Server.get_current().to_json()

        self.client = self.app.test_client()

    def tearDown(self) -> None:
        with self.app_join.app_context():
            db.session.remove()
            db.drop_all()
        with self.app.app_context():
            db.session.remove()
            db.drop_all()

    @patch('dimensigon.domain.entities.route.check_host')
    @patch('dimensigon.domain.entities.get_now')
    # @patch('dimensigon.rsa.newkeys')
    @aioresponses()
    def test_join_command(self, mock_now, mock_check, m):
        mock_now.return_value = dt.datetime(2019, 5, 1, tzinfo=dt.timezone.utc)

        mock_check.return_value = True

        def callback_client(method, client, url, **kwargs):
            kwargs.pop('allow_redirects')

            func = getattr(client, method.lower())
            r = func(url.path, headers=kwargs['headers'], json=kwargs['json'])

            return CallbackResult(method.upper(), status=r.status_code, body=r.data, content_type=r.content_type,
                                  headers=r.headers)

        m.post(re.compile('^https?://node1.*'),
               callback=functools.partial(callback_client, 'POST', self.client), repeat=True)

        def request_callback(request, client):
            method_func = getattr(client, request.method.lower())
            resp = method_func(request.path_url, data=request.body, headers=dict(request.headers))

            return resp.status_code, resp.headers, resp.data

        with responses.RequestsMock() as rsps:
            rsps.add_callback(responses.POST,
                              re.compile('^https?://node1.*'),
                              callback=partial(request_callback, client=self.client))
            rsps.add_callback(responses.GET,
                              re.compile('^https?://node1.*'),
                              callback=partial(request_callback, client=self.client))

            with self.app.app_context():
                runner = self.app.test_cli_runner()
                result = runner.invoke(generate_token, [], catch_exceptions=False)
                token = result.stdout.strip()

            runner = self.app_join.test_cli_runner()

            result = runner.invoke(join, ['--no-ssl', 'node1:5000', token], catch_exceptions=False)

        with self.app.app_context():
            # check if new server created
            server_new = Server.query.get(self.server_new.get('id'))
            self.assertEqual(server_new.last_modified_at, dt.datetime(2019, 5, 1, tzinfo=dt.timezone.utc))

        with self.app_join.app_context():
            self.assertDictEqual(self.dimension, Dimension.get_current().to_json())
            s = Server.query.get(self.server_node1.get('id'))
            self.assertEqual(0, s.route.cost)
            self.assertListEqual([('Gate',), ('Server',), ('User',)],
                                 db.session.query(Catalog.entity).order_by('entity').all())
            self.assertEqual(dt.datetime(2019, 5, 1, tzinfo=dt.timezone.utc),
                             Catalog.query.filter_by(entity='Server').one().last_modified_at)



import logging
import os
import time
import typing as t

from sqlalchemy import create_engine, text
from sqlalchemy.exc import InternalError, OperationalError, SQLAlchemyError
from sqlalchemy.orm import sessionmaker, scoped_session
from sqlalchemy.sql.ddl import CreateTable

from dimensigon import defaults
from dimensigon.core import Dimensigon
from dimensigon.domain.entities import SCHEMA_VERSION, SchemaChanges
from dimensigon.utils.helpers import session_scope
from dimensigon.web import db

_LOGGER = logging.getLogger('dimensigon.db')

PROGRESS_FILE = ".migration_progress"


def setup_db(dm: Dimensigon):
    """Ensure database is ready to fly."""

    dm.engine = create_engine(dm.config.db_uri)
    if not os.path.exists(dm.config.db_uri[len(defaults.DB_PREFIX):]):
        _LOGGER.info(f"Creating database {dm.config.db_uri[len(defaults.DB_PREFIX):]}")
        db.Model.metadata.create_all(dm.engine)
        time.sleep(5)
    else:
        _LOGGER.debug(f"Database {dm.config.db_uri[len(defaults.DB_PREFIX):]} already exists")
    dm.get_session = scoped_session(sessionmaker(bind=dm.engine))
    migrate_schema(dm)

    populate_initial_data(dm)


def migrate_schema(dm: Dimensigon):
    progress_path = dm.config.path(PROGRESS_FILE)
    with session_scope(session=dm.get_session()) as session:

        res = (
            session.query(SchemaChanges)
                .order_by(SchemaChanges.change_id.desc())
                .first()
        )
        current_version = getattr(res, "schema_version", None)

        if current_version is None:
            # first time running database
            sc = SchemaChanges(schema_version=SCHEMA_VERSION)
            session.add(sc)
            current_version = SCHEMA_VERSION

        if current_version == SCHEMA_VERSION:
            # Clean up if old migration left file
            if os.path.isfile(progress_path):
                _LOGGER.warning("Found existing migration file, cleaning up")
                os.remove(dm.config.path(PROGRESS_FILE))
            return

        with open(progress_path, "w"):
            pass

        _LOGGER.warning(
            "Database is about to upgrade. Schema version: %s", current_version
        )

        try:
            for version in range(current_version, SCHEMA_VERSION):
                new_version = version + 1
                _LOGGER.info("Upgrading recorder db schema to version %s", new_version)
                _apply_update(dm.engine, new_version, current_version)
                session.add(SchemaChanges(schema_version=new_version))

                _LOGGER.info("Upgrade to version %s done", new_version)
        finally:
            os.remove(dm.config.path(PROGRESS_FILE))


def populate_initial_data(dm: Dimensigon):
    from dimensigon.domain.entities import ActionTemplate, Locker, Server, User, Parameter

    with session_scope(session=dm.get_session()) as session:
        gates = dm.config.http_conf.get('binds', None)

        Server.set_initial(session, gates)

        Locker.set_initial(session)
        ActionTemplate.set_initial(session)
        User.set_initial(session)
        Parameter.set_initial(session)


def _add_columns(engine, table_name, columns_def):
    """Add columns to a table."""
    _LOGGER.warning(
        "Adding columns %s to table %s. Note: this can take several "
        "minutes on large databases and slow computers. Please "
        "be patient!",
        ", ".join(column.split(" ")[0] for column in columns_def),
        table_name,
    )

    columns_def = [f"ADD {col_def}" for col_def in columns_def]

    # try:
    #     engine.execute(
    #         text(
    #             "ALTER TABLE {table} {columns_def}".format(
    #                 table=table_name, columns_def=", ".join(columns_def)
    #             )
    #         )
    #     )
    #     return
    # except (InternalError, OperationalError):
    #     # Some engines support adding all columns at once,
    #     # this error is when they don't
    #     _LOGGER.info("Unable to use quick column add. Adding 1 by 1")

    with engine.connect() as connection:
        for column_def in columns_def:
            try:
                connection.execute(
                    text(
                        "ALTER TABLE {table} {column_def}".format(
                            table=table_name, column_def=column_def
                        )
                    )
                )
            except (InternalError, OperationalError) as err:
                if "duplicate" not in str(err).lower():
                    raise

                _LOGGER.warning(
                    "Column %s already exists on %s, continuing",
                    column_def.split(" ")[1],
                    table_name,
                )


def _rename_columns(engine, table_name, column_renames: t.List[t.Tuple[str, str]]):
    temp_table_name = table_name + '_temp'
    table = db.Model.metadata.tables[table_name]
    tmp_ddl = CreateTable(table).compile(engine).string.replace(table_name, temp_table_name)

    map2old = {t[1]: t[0] for t in column_renames}
    new_column_table = table.columns.keys()

    old_column_table = [map2old.get(item, item) for item in new_column_table]

    with engine.connect() as connection:
        old_c_s = [f'"{c}"' for c in old_column_table]
        new_c_s = [f'"{c}"' for c in new_column_table]
        connection.execute(tmp_ddl)
        try:
            connection.execute(f'INSERT INTO {temp_table_name}({", ".join(new_c_s)}) '
                               f'SELECT {", ".join(old_c_s)} FROM {table_name}')

        except:
            connection.execute(f"DROP TABLE {temp_table_name}")
            raise
        else:
            connection.execute(f"DROP TABLE {table_name}")
            connection.execute(f"ALTER TABLE {temp_table_name} RENAME TO {table_name}")


def _delete_columns(engine, table_name, column_deletes: t.List[str]):
    temp_table_name = table_name + '_temp'
    table = db.Model.metadata.tables[table_name]
    tmp_ddl = CreateTable(table).compile(engine).string.replace(table_name, temp_table_name)

    new_column_table = [k for k in table.columns.keys() if k not in column_deletes]

    with engine.connect() as connection:
        new_c_s = [f'"{c}"' for c in new_column_table]
        connection.execute(tmp_ddl)
        try:
            connection.execute(f'INSERT INTO {temp_table_name}({", ".join(new_c_s)}) '
                               f'SELECT {", ".join(new_c_s)} FROM {table_name}')

        except:
            connection.execute(f"DROP TABLE {temp_table_name}")
            raise
        else:
            connection.execute(f"DROP TABLE {table_name}")
            connection.execute(f"ALTER TABLE {temp_table_name} RENAME TO {table_name}")


def _rename_and_delete_columns(engine, table_name,
                               column_renames: t.List[t.Tuple[str, str]],
                               column_deletes: t.List[str]):
    temp_table_name = table_name + '_temp'
    table = db.Model.metadata.tables[table_name]
    tmp_ddl = CreateTable(table).compile(engine).string.replace(table_name, temp_table_name)

    map2old = {t[1]: t[0] for t in column_renames}
    new_column_table = [k for k in table.columns.keys() if k not in column_deletes]

    old_column_table = [map2old.get(item, item) for item in new_column_table]

    with engine.connect() as connection:
        old_c_s = [f'"{c}"' for c in old_column_table]
        new_c_s = [f'"{c}"' for c in new_column_table]
        connection.execute(tmp_ddl)
        try:
            connection.execute(f'INSERT INTO {temp_table_name}({", ".join(new_c_s)}) '
                               f'SELECT {", ".join(old_c_s)} FROM {table_name}')

        except:
            connection.execute(f"DROP TABLE {temp_table_name}")
            raise
        else:
            connection.execute(f"DROP TABLE {table_name}")
            connection.execute(f"ALTER TABLE {temp_table_name} RENAME TO {table_name}")


def _create_table(engine, table_name):
    table = db.Model.metadata.tables[table_name]
    ddl = CreateTable(table).compile(engine).string
    with engine.connect() as connection:
        connection.execute(ddl)

def _create_index(engine, table_name, index_name):
    """Create an index for the specified table.
    The index name should match the name given for the index
    within the table definition described in the models
    """
    table = db.Model.metadata.tables[table_name]
    _LOGGER.debug("Looking up index %s for table %s", index_name, table_name)
    # Look up the index object by name from the table is the models
    index_list = [idx for idx in table.indexes if idx.name == index_name]
    if not index_list:
        _LOGGER.debug("The index %s no longer exists", index_name)
        return
    index = index_list[0]
    _LOGGER.debug("Creating %s index", index_name)
    _LOGGER.warning(
        "Adding index `%s` to database. Note: this can take several "
        "minutes on large databases and slow computers. Please "
        "be patient!",
        index_name,
    )
    try:
        index.create(engine)
    except OperationalError as err:
        lower_err_str = str(err).lower()

        if "already exists" not in lower_err_str and "duplicate" not in lower_err_str:
            raise

        _LOGGER.warning(
            "Index %s already exists on %s, continuing", index_name, table_name
        )
    except InternalError as err:
        if "duplicate" not in str(err).lower():
            raise

        _LOGGER.warning(
            "Index %s already exists on %s, continuing", index_name, table_name
        )

    _LOGGER.debug("Finished creating %s", index_name)


def _drop_index(engine, table_name, index_name):
    """Drop an index from a specified table.
    There is no universal way to do something like `DROP INDEX IF EXISTS`
    so we will simply execute the DROP command and ignore any exceptions
    WARNING: Due to some engines (MySQL at least) being unable to use bind
    parameters in a DROP INDEX statement (at least via SQLAlchemy), the query
    string here is generated from the method parameters without sanitizing.
    DO NOT USE THIS FUNCTION IN ANY OPERATION THAT TAKES USER INPUT.
    """
    _LOGGER.debug("Dropping index %s from table %s", index_name, table_name)
    success = False

    # Engines like DB2/Oracle
    try:
        engine.execute(text(f"DROP INDEX {index_name}"))
    except SQLAlchemyError:
        pass
    else:
        success = True

    # Engines like SQLite, SQL Server
    if not success:
        try:
            engine.execute(
                text(
                    "DROP INDEX {table}.{index}".format(
                        index=index_name, table=table_name
                    )
                )
            )
        except SQLAlchemyError:
            pass
        else:
            success = True

    if not success:
        # Engines like MySQL, MS Access
        try:
            engine.execute(
                text(
                    "DROP INDEX {index} ON {table}".format(
                        index=index_name, table=table_name
                    )
                )
            )
        except SQLAlchemyError:
            pass
        else:
            success = True

    if success:
        _LOGGER.debug(
            "Finished dropping index %s from table %s", index_name, table_name
        )
    else:
        if index_name == "ix_states_context_parent_id":
            # Was only there on nightly so we do not want
            # to generate log noise or issues about it.
            return

        _LOGGER.warning(
            "Failed to drop index %s from table %s. Schema "
            "Migration will continue; this is not a "
            "critical operation",
            index_name,
            table_name,
        )

def _apply_update(engine, new_version, old_version):
    if new_version == 2:
        _rename_columns(engine, 'D_server', [('unreachable', 'alive')])
    elif new_version == 3:
        _add_columns(engine, 'D_server', ['created_on DATETIME'])
        _delete_columns(engine, 'D_server', ['alive'])
        with engine.connect() as connection:
            date = defaults.INITIAL_DATEMARK.strftime('%Y-%m-%d %H:%M:%S.%f')
            connection.execute(
                f"UPDATE D_server SET created_on = '{date}' WHERE created_on IS NULL")
    elif new_version == 4:
        _create_table(engine, 'L_parameter')
    elif new_version == 5:
        _add_columns(engine, 'D_gate', ['deleted BOOLEAN'])
        with engine.connect() as connection:
            connection.execute(
                f"UPDATE D_gate SET deleted = 0")
            connection.execute(
                f"UPDATE D_gate SET deleted = 1 WHERE EXISTS(SELECT * FROM D_server WHERE  D_server.id == D_gate.server_id and D_server.deleted == 1)")



import abc
import functools
import inspect
import random
import string
import uuid
from datetime import datetime

from sqlalchemy import Column, Boolean

from dimensigon import defaults
from dimensigon.utils.typos import UUID, UtcDateTime
from dimensigon.web.helpers import QueryWithSoftDelete


class JSONEntity:

    @abc.abstractmethod
    def to_json(self):
        ...

    @classmethod
    @abc.abstractmethod
    def from_json(cls, kwargs):
        ...


class SoftDeleteMixin:
    __prefix__ = '_old_'
    deleted = Column(Boolean(), default=False)

    def __new__(cls, *args, **kwargs):
        obj = super(SoftDeleteMixin, cls).__new__(cls)
        obj.query_class = QueryWithSoftDelete
        if hasattr(obj, 'to_json') and callable(getattr(obj, 'to_json')):
            obj.to_json = SoftDeleteMixin.wrapper_to_json(obj, obj.to_json)

        return obj

    def __init__(self, deleted=False, **kwargs):
        self.deleted = deleted
        for attr, value in kwargs.items():
            if attr.startswith(self.__prefix__):
                setattr(self, attr, kwargs.get(attr, None))

    def delete(self):
        if not self.deleted:
            self.deleted = True
            for attr in [attr for attr, value in inspect.getmembers(self) if attr.startswith(self.__prefix__)]:
                original_attr = attr.lstrip(self.__prefix__)
                setattr(self, attr, getattr(self, original_attr))
                setattr(self, original_attr,
                        ''.join(random.choices(string.digits + string.ascii_letters + string.punctuation, k=10)))


    def wrapper_to_json(self, func):

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            no_delete = kwargs.pop('no_delete', False)
            dto = func(*args, **kwargs)
            if not no_delete:
                dto.update({'deleted': self.deleted})
                for attr in [attr for attr, value in inspect.getmembers(self) if attr.startswith(self.__prefix__)]:
                    dto.update({attr: getattr(self, attr)})
            return dto

        return wrapper


class DistributedEntityMixin(JSONEntity):
    order = None
    last_modified_at = Column(UtcDateTime(), nullable=False)

    def __init__(self, **kwargs):
        self.last_modified_at = kwargs.pop('last_modified_at', None)

    def to_json(self):
        try:
            return dict(last_modified_at=self.last_modified_at.strftime(defaults.DATEMARK_FORMAT))
        except AttributeError:
            return dict()

    @classmethod
    def from_json(cls, kwargs):
        if 'last_modified_at' in kwargs:
            last_modified_at = kwargs.pop('last_modified_at')
            last_modified_at = datetime.strptime(last_modified_at, defaults.DATEMARK_FORMAT)
            kwargs.update(last_modified_at=last_modified_at)


class UUIDEntityMixin:
    id = Column(UUID, primary_key=True, default=lambda: str(uuid.uuid4()))

    def __init__(self, **kwargs):
        if 'id' in kwargs:
            self.id = str(kwargs['id']).lower()
        else:
            self.id = str(uuid.uuid4())


class UUIDistributedEntityMixin(UUIDEntityMixin, DistributedEntityMixin):

    def __init__(self, **kwargs):
        UUIDEntityMixin.__init__(self, **kwargs)
        DistributedEntityMixin.__init__(self, **kwargs)

    def to_json(self):
        data = super().to_json()
        if self.id:
            data['id'] = str(self.id)
        return data

    @classmethod
    @abc.abstractmethod
    def from_json(cls, kwargs):
        super().from_json(kwargs)
        try:
            o = cls.query.get(kwargs.get('id'))
        except RuntimeError as e:
            o = None
        if o:
            for k, v in kwargs.items():
                if getattr(o, k) != v:
                    setattr(o, k, v)
            return o
        else:
            return cls(**kwargs)



class EntityReprMixin:
    id = None

    def __repr__(self):
        if self.id:
            return f'<{self.__class__.__name__} {self.id}>'
        else:
            return f'<{self.__class__.__name__} (transient {id(self)})>'

    def __str__(self):
        return self.__repr__()



import datetime as dt
import itertools
import os
import socket

import netifaces

# Global defaults
CONFIG_DIR_NAME = ".dimensigon"
HOME = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))


ips = list(filter(lambda x: x != '127.0.0.1', itertools.chain(
    *[[ip['addr'] for ip in netifaces.ifaddresses(iface).get(netifaces.AF_INET, [])] for iface in
      netifaces.interfaces()])))

# Gunicorn Defaults
DEFAULT_PORT = 5000
HOSTNAME = socket.gethostname()
GUNICORN_CONF_FILE = '_gunicorn.conf.py'
PID_FILE = 'dimensigon.pid'
PROC_NAME = 'dimensigon'
ACCESS_LOGFILE = 'access.log'
ERROR_LOGFILE = 'dimensigon.log'
SSL_DIR = '.ssl'
KEY_FILE = 'key.pem'
CERT_FILE = 'cert.pem'

# Database
DB_PREFIX = 'sqlite:///'
DEFAULT_DB_URL = f"{DB_PREFIX}{{db_file}}"
DEFAULT_DB_FILE = "dimensigon.db"

# Dimensigon Defaults
MAX_WAITING_TIME = 300  # time in seconds waiting tasks to finish
MAX_TIME_WAITING_SERVERS = 600  # max time waiting for servers to be created
JOIN_TOKEN_EXPIRE_TIME = 15  # join token expire time in minutes
REFRESH_PERIOD = 5  # Run route table, catalog and cluster refresh every minutes
COMA_NODE_FACTOR = 2  # multiply factor to decide if a node is in coma state. In time: COMA_NODE_FACTOR x REFRESH_PERIOD

# quorum algorithm
ADULT_NODES = dt.timedelta(hours=24)  # age of a node to be selectable for the quorum
MAJORITY = 0.51  # ratio of servers to be selectable from old population in order to perform a lock

TIMEOUT_REQUEST = 60
TIMEOUT_PREVENTING_LOCK = 60  # max time in seconds locker will be in PREVENTING_LOCK before returning to UNLOCK
TIMEOUT_ORCHESTRATION = 1800  # max time waiting for an orchestration to finish
TIMEOUT_COMMAND = 20  # max time waiting for a command execution
TIMEOUT_LOCK_REQUEST = 60  # timeout on lock/unlock/prevent_lock HTTP request


CHUNK_SIZE = 2*1024  # in MB
MAX_SENDERS = 4

MIN_SERVERS_QUORUM = 5  # minimum servers to run quorum algorithm
INITIAL_DATEMARK = dt.datetime(2019, 4, 1, tzinfo=dt.timezone.utc)

DATETIME_FORMAT = "%m/%d/%Y, %H:%M:%S.%f %z"
DATEMARK_FORMAT = "%Y%m%d.%H%M%S.%f%z"

SOFTWARE_REPO = 'software'
DIMENSIGON_DIR = os.path.join(SOFTWARE_REPO, 'dimensigon')
LOG_FOLDER = 'logs'
LOG_SENDER_REPO = 'logfed'





import datetime as dt
import unittest

from dimensigon import defaults
from dimensigon.domain.entities.parameter import Parameter
from dimensigon.utils.helpers import get_now
from dimensigon.web import create_app, db


class TestParameter(unittest.TestCase):

    def setUp(self):
        """Create and configure a new app instance for each test."""
        # create the app with common test config
        self.app = create_app('test')
        self.app_context = self.app.app_context()
        self.app_context.push()
        db.create_all()

    def tearDown(self) -> None:
        db.session.remove()
        db.drop_all()
        self.app_context.pop()

    def test_set_and_get(self):
        p = Parameter('data', dump=lambda x: x.strftime(defaults.DATETIME_FORMAT),
                            load=lambda x: dt.datetime.strptime(x, defaults.DATETIME_FORMAT))
        db.session.add(p)
        now = get_now()
        Parameter.set('data', now)

        p = Parameter.query.get('data')
        self.assertEqual(now.strftime(defaults.DATETIME_FORMAT), p.value)

        p = Parameter('integer', load=int, dump=str)
        db.session.add(p)
        Parameter.set('integer', 5)
        p = Parameter.query.get('integer')
        self.assertEqual('5', p.value)

        p = Parameter('name')
        db.session.add(p)
        Parameter.set('name', 'Joan')
        p = Parameter.query.get('name')
        self.assertEqual('Joan', p.value)


        # test get
        self.assertEqual(now, Parameter.get('data'))
        self.assertEqual(5, Parameter.get('integer'))
        self.assertEqual('Joan', Parameter.get('name'))

        p = Parameter('none')
        db.session.add(p)

        self.assertIsNone(Parameter.get('none'))
        self.assertEqual('default', Parameter.get('none', 'default'))


import argparse as argparse
import functools
import json
import os
import uuid

import requests
from prompt_toolkit.completion import merge_completers

from dimensigon import defaults
from dimensigon.domain.entities import ActionType, Scope
from dimensigon.domain.entities.transfer import Status
from dimensigon.dshell import environ
from dimensigon.dshell import environ as env
from dimensigon.dshell.argparse_raise import ParamAction, ExtendAction
from dimensigon.dshell.bootstrap import save_config_file
from dimensigon.dshell.completer import *
from dimensigon.dshell.helpers import name2id, exit_dshell
from dimensigon.dshell.output import dprint
from dimensigon.dshell.prompts.action_template import subprompt as action_prompt
from dimensigon.dshell.prompts.command import subprompt as command_prompt
from dimensigon.dshell.prompts.orchestration import subprompt as orch_prompt
from dimensigon.dshell.utils import clean_none
from dimensigon.utils.helpers import get_now, is_valid_uuid


def status(node, detail=False):
    view_data = {}
    if not detail:
        view_data.update(params='human')
    if not node:
        resp = ntwrk.get('root.healthcheck', view_data=view_data)
        dprint(resp)
    else:
        for n in node:
            dprint(f"### {n}:") if len(node) > 1 else None
            if not is_valid_uuid(n):
                node_id = name2id('api_1_0.serverlist', n)
            else:
                node_id = n
            resp = ntwrk.get('root.healthcheck', view_data=view_data, headers={'D-Destination': node_id})
            dprint(resp)


def ping(node):
    for n in node:
        dprint(f"### {n}:") if len(node) > 1 else None
        if not is_valid_uuid(n):
            node_id = name2id('api_1_0.serverlist', n)
        else:
            node_id = n
        resp = ntwrk.post('root.ping', headers={'D-Destination': node_id},
                          json={'start_time': get_now().strftime(defaults.DATETIME_FORMAT)})
        dprint(resp)


def manager_locker_show(node):
    for n in node:
        dprint(f"### {n}:") if len(node) > 1 else None
        if not is_valid_uuid(n):
            node_id = name2id('api_1_0.serverlist', n)
        else:
            node_id = n
        resp = ntwrk.get('api_1_0.locker', headers={'D-Destination': node_id})
        dprint(resp)


def manager_locker_unlock(scope, node):
    for n in node:
        dprint(f"### {n}:") if len(node) > 1 else None
        if not is_valid_uuid(n):
            node_id = name2id('api_1_0.serverlist', n)
        else:
            node_id = n
        resp = ntwrk.post('api_1_0.locker_unlock',
                          json={'scope': scope, 'applicant': "", 'force': True},
                          headers={'D-Destination': node_id})
        dprint(resp)


def server_list(name=None, iden=None, detail=None, like=None):
    kwargs = dict(verify=environ.get('SSL_VERIFY'))
    view_data = dict()
    if name is not None:
        view_data.update({'filter[name]': name})
    if iden:
        view_data.update({'filter[id]': iden})
    if detail:
        view_data.update(params='gates')
    resp = ntwrk.get('api_1_0.serverlist', view_data=view_data, **kwargs)
    if resp.code == 200:
        data = resp.msg or []
        # post process
        filtered_data = []
        if like:
            for server in data:
                if like in server.get('name'):
                    filtered_data.append(server)
        else:
            filtered_data = data
        dprint(filtered_data)
    else:
        dprint(resp)


def server_delete(server_ids):
    converted_ids = []
    for server_id in server_ids:
        try:
            uuid.UUID(server_id)
        except Exception:
            not_an_uuid = True
        else:
            not_an_uuid = False
        if not_an_uuid:
            server_id = name2id('api_1_0.serverlist', server_id)
        converted_ids.append(server_id)

    if converted_ids:
        resp = ntwrk.delete('api_1_0.serverlist', json={'server_ids': converted_ids})
        if resp.ok:
            dprint(f"Server{'s' if len(converted_ids) > 1 else ''} removed successfully")
        else:
            dprint(resp)
    else:
        dprint("No server no delete")


def server_routes(node, refresh=False):
    resp = None
    if not node:
        node = ['localhost']
    for n in node:
        dprint(f"### {n}:") if len(node) > 1 else None
        if n == 'localhost':
            node_id = None
        elif not is_valid_uuid(n):
            node_id = name2id('api_1_0.serverlist', n)
        else:
            node_id = n

        kwargs = {}
        if node_id:
            kwargs.update(headers={'D-Destination': node_id})

        if refresh:
            resp = ntwrk.post('api_1_0.routes', json={"discover_new_neighbours": True,
                                                      "check_current_neighbours": True}, **kwargs)
        if resp is None or (resp and resp.ok):
            resp = ntwrk.get('api_1_0.routes', view_data={'params': 'human'}, **kwargs)
        dprint(resp)


def orch_list(**params):
    kwargs = dict(verify=environ.get('SSL_VERIFY'))
    view_data = dict()
    if params.get('id', None):
        view_data.update({'filter[id]': params['id']})
    if params.get('name'):
        view_data.update({'filter[name]': params['name']})
    if params.get('version', None):
        view_data.update({'filter[version]': str(params['version'])})
    if params.get('detail', None):
        view_data.update({'params': ['steps', 'vars', 'target', 'action', 'human']})
    else:
        view_data.update({'params': ['vars', 'target', 'human']})
    resp = ntwrk.get('api_1_0.orchestrationlist', view_data=view_data, **kwargs)
    if resp.code == 200:
        data = resp.msg or []
        # post process
        filtered_data = []
        if params.get('like', None):
            for server in data:
                if params.get('like') in server.get('name'):
                    filtered_data.append(server)
        else:
            filtered_data = data
        dprint(filtered_data)
    else:
        dprint(resp)


def orch_create(**params):
    orch_prompt({'name': params['name'], 'steps': []},
                changed=True,
                ask_all=params['prompt'], parent_prompt='Δ ')


def orch_copy(**params):
    resp = ntwrk.get('api_1_0.orchestrationresource',
                     view_data={'orchestration_id': params.get('orchestration_id')})

    if resp.ok:
        orch = resp.msg
        orch.pop('target', None)
        orch.pop('params', None)
        orch.pop('id', None)
        orch.pop('version', None)
        orch.pop('last_modified_at', None)
        resp = ntwrk.get('api_1_0.steplist',
                         view_data={'filter[orchestration_id]': params.get('orchestration_id')})
        if resp.ok:
            orch['steps'] = resp.msg

            # remove unwanted parameters
            for s in orch['steps']:
                s.pop('last_modified_at', None)
                s.pop('created_on', None)
                s.pop('orchestration_id', None)

            orch_prompt(entity=orch, parent_prompt='Δ ')


def orch_load(file):
    try:
        orch = json.load(file)
    except Exception as e:
        print(e)
    else:
        orch_prompt(orch, parent_prompt='Δ ')


def orch_run(orchestration_id, **params):
    if not params['hosts']:
        print('No target specified')
        return
    resp = ntwrk.post('api_1_0.launch_orchestration',
                      view_data={'orchestration_id': orchestration_id, 'params': 'human'}, json=params)
    dprint(resp)


def action_list(iden=None, name=None, version=None, like: str = None, last: int = None):
    kwargs = dict(verify=environ.get('SSL_VERIFY'))
    view_data = dict()
    if iden:
        view_data.update({'filter[id]': iden})
    if name:
        view_data.update({'filter[name]': name})
    if version:
        view_data.update({'filter[version]': version})
    resp = ntwrk.get('api_1_0.actiontemplatelist', view_data=view_data, **kwargs)
    if resp.code == 200:
        data = resp.msg or []
        # post process
        filtered_data = []
        if like:
            for server in data:
                if like in server.get('name'):
                    filtered_data.append(server)
        else:
            filtered_data = data
        dprint(filtered_data)
    else:
        dprint(resp)


def action_create(name, action_type, prompt):
    action_prompt({'name': name, 'action_type': action_type}, changed=True, ask_all=prompt, parent_prompt='Δ ')


def action_copy(**params):
    resp = ntwrk.get('api_1_0.actiontemplateresource',
                     view_data={'action_template_id': params.get('action_template_id')})

    if resp.ok:
        action = resp.msg
        action.pop('id', None)
        action.pop('version', None)
        action.pop('last_modified_at', None)
        # clean empty fields
        for k, v in dict(action).items():
            if v is None:
                action.pop(k, None)

        action_prompt(entity=action, parent_prompt='Δ ')


def action_load(file):
    try:
        action = json.load(file)
    except Exception as e:
        print(e)
    else:
        action_prompt(action, parent_prompt='Δ ')


def software_add(name, version, file, family=None):
    kwargs = dict(verify=environ.get('SSL_VERIFY'))
    json_data = dict(name=name, version=version, file=file)
    if family is not None:
        json_data.update(family=family)
    resp = ntwrk.post(view='api_1_0.softwarelist', json=json_data, **kwargs)
    dprint(resp)


def software_send(dest_server_id, software_id=None, software=None, version=None, file=None, dest_path=None,
                  background=True, force=False):
    kwargs = dict(verify=environ.get('SSL_VERIFY'))
    json_data = dict(dest_server_id=dest_server_id, dest_path=dest_path, force=force)
    if software_id:
        json_data.update(software_id=software_id)
    elif file:
        json_data.update(file=file)
    else:
        json_data.update(software=software, version=version)

    if not background:
        json_data.update(include_transfer_data=True)
    resp = ntwrk.post('api_1_0.send', json=json_data, **kwargs)
    dprint(resp)


def software_list(name=None, version=None, detail=None, like=None):
    kwargs = dict(verify=environ.get('SSL_VERIFY'))
    view_data = dict()
    if name:
        view_data.update({'filter[name]': name})
    if version:
        view_data.update({'filter[version]': version})
    if detail:
        view_data.update(params='servers')
    resp = ntwrk.get('api_1_0.softwarelist', view_data=view_data, **kwargs)
    if resp.ok:
        data = resp.msg or []
        # post process
        filtered_data = []
        if like:
            for server in data:
                if like in server.get('name'):
                    filtered_data.append(server)
        else:
            filtered_data = data
        dprint(filtered_data)
    else:
        dprint(resp)


def transfer_list(iden=None, status=None, like=None):
    kwargs = dict(verify=environ.get('SSL_VERIFY'))
    view_data = dict()
    if iden:
        view_data.update({'filter[id]': iden})
    if status:
        view_data.update({'filter[status]': ','.join(status)})
    resp = ntwrk.get('api_1_0.transferlist', view_data=view_data, **kwargs)
    if resp.ok:
        data = resp.msg or []
        # post process
        filtered_data = []
        if like:
            for server in data:
                if like in server.get('name'):
                    filtered_data.append(server)
        else:
            filtered_data = data
        dprint(filtered_data)
    else:
        dprint(resp)


def transfer_cancel(transfer_id):
    kwargs = dict(verify=environ.get('SSL_VERIFY'))
    data = {'status': 'CANCELED'}
    resp = ntwrk.patch('api_1_0.transferresource', view_data={'transfer_id': transfer_id}, json=data, **kwargs)
    dprint(resp)


def exec_list(orch=None, server=None, execution_id=None, last=None, asc=None, detail=None):
    kwargs = dict(verify=environ.get('SSL_VERIFY'))
    view_data = dict()
    view = 'api_1_0.orchexecutionlist'

    if orch:
        view_data.update({'filter[orchestration_id]': orch})

    if server:
        view_data.update({'filter[server_id]': server})

    if execution_id:
        view_data.update({'filter[id]': execution_id})

    # url_parameters
    if detail:
        view_data.update({'params': ['steps', 'human']})
    else:
        view_data.update({'params': ['human']})

    resp = ntwrk.get(view, view_data=view_data, **kwargs)
    if resp.ok:
        data = resp.msg or []
        # post process
        if asc:
            data.sort(key=lambda x: x.get('start_time'))
        else:
            data.sort(key=lambda x: x.get('start_time'), reverse=True)

        if last:
            dprint(data[:last])
        else:
            dprint(data)
    else:
        dprint(resp)


def cmd(command, target, timeout=None, input=None, shell=None):
    if shell:
        command_prompt({'target': target}, ask_all=False if target else True, parent_prompt='Δ')
    else:
        if isinstance(command, list):
            command = ' '.join(command)
        if not command:
            raise ValueError('a command must be specified')
        if not target:
            raise ValueError('target must be specified')
        data = {'command': command, 'target': target}
        if timeout:
            data.update(timeout=timeout)
        if input:
            data.update(input=input.replace('\\n', '\n').replace('\\t', '\t'))
        resp = ntwrk.post('api_1_0.launch_command', view_data={'params': 'human'}, json=data)
        dprint(resp)


def logfed_list():
    kwargs = dict(verify=environ.get('SSL_VERIFY'))
    resp = ntwrk.get('api_1_0.loglist', view_data={'params': 'human'}, **kwargs)
    dprint(resp.msg if resp.code else str(resp.exception) if str(
        resp.exception) else resp.exception.__class__.__name__)


def logfed_subscribe(src_server_id, target, dst_server_id, include=None, exclude=None, dest_folder=None,
                     recursive=None, mode=None):
    if not is_valid_uuid(src_server_id):
        _src_server_id = name2id('api_1_0.serverlist', src_server_id)
    else:
        _src_server_id = src_server_id
    if not is_valid_uuid(dst_server_id):
        _dst_server_id = name2id('api_1_0.serverlist', dst_server_id)
    else:
        _dst_server_id = dst_server_id
    kwargs = dict(verify=environ.get('SSL_VERIFY'))
    json_data = dict(src_server_id=_src_server_id, target=target, dst_server_id=_dst_server_id, include=include,
                     exclude=exclude, dest_folder=dest_folder, recursive=recursive, mode=mode)
    json_data = clean_none(json_data)
    resp = ntwrk.post('api_1_0.loglist', json=json_data, **kwargs)
    dprint(resp)


def logfed_unsubscribe(log_id):
    kwargs = dict(verify=environ.get('SSL_VERIFY'))
    resp = ntwrk.delete('api_1_0.logresource', {'log_id': log_id}, **kwargs)
    dprint(resp)


def manager_catalog_refresh():
    resp = ntwrk.post('api_1_0.catalog_update')
    dprint(resp)


def manager_locker_ignore(ignore: bool, server_ids):
    resp = ntwrk.post('api_1_0.manager_server_ignore_lock', json={'server_ids': server_ids, 'ignore_on_lock': ignore})
    dprint(resp)



# def locker_lock(**params):
#     pass
#
# def locker_unlock(**params):
#     if params['servers'] is None:
#         resp = get('api_1_0.serverlist')
#         resp.raise_if_not_ok()
#         servers = [s.id for s in resp.msg]
#     else:
#         servers = params['servers']
#
#     for server in servers:
#         resp = post('api_1_0.locker_unlock')

def manager_token(raw=False, expires_time=None):
    resp = ntwrk.get('api_1_0.join_token', view_data=dict(expires_time=expires_time))
    if not raw:
        dprint(resp)
    else:
        if resp.ok:
            dprint(resp.msg['token'])
        else:
            dprint(resp)

def manager_login(username=None, password=None):
    try:
        ntwrk.login(username, password)
    except requests.exceptions.ConnectionError as e:
        dprint(f"Unable to contact with {environ.get('SCHEME')}://{environ.get('SERVER')}:{environ.get('PORT')}/")
    except Exception as e:
        dprint(str(e))


def manager_save_login():
    save_config_file(os.path.expanduser(env.get('CONFIG_FILE', None)), username=env._username, token=env._refresh_token, server=env.get('SERVER'), port=env.get('PORT'))

def logging_cmd(logger, level):
    logger = logging.getLogger(logger)
    if logger:
        logger.setLevel(level)
    else:
        dprint(f"Logger '{logger}' does not exist")


def env_list():
    for k, v in environ._environ.items():
        dprint(f"{k}={v}")


def env_get(key):
    dprint(f"{environ.get(key, None)}")


def env_set(key, value):
    if isinstance(value, list) and len(value) == 0:
        if '=' in key:
            key, value = key.split('=')
            environ.set(key, value)
        else:
            dprint('invalid value')
    else:
        environ.set(key, ' '.join(value))


nested_dict = {
    'status': [{'argument': 'node', 'nargs': '*', 'completer': server_completer},
               {'argument': '--detail', 'action': 'store_true'}, status],
    'ping': [{'argument': 'node', 'nargs': '+', 'completer': server_completer}, ping],
    'server': {
        'list': [{'argument': '--detail', 'action': 'store_true'},
                 [{'argument': '--like'},
                  {'argument': '--name', 'completer': server_name_completer},
                  {'argument': '--id', 'dest': 'iden', 'completer': server_completer},
                  # {'argument': '--last', 'action': 'store', 'type': int}
                  ],
                 server_list
                 ],
        'delete': [
            {'argument': 'server_ids', 'metavar': 'NODE', 'nargs': '+', 'completer': server_completer,
             'help': 'Node to be deleted'},
            server_delete],
        'routes': [{'argument': 'node', 'nargs': '*', 'completer': server_completer},
                   {'argument': '--refresh', 'action': 'store_true'},
                   server_routes],
    },
    'orch': {
        'list': [{'argument': '--version', 'action': 'store', 'type': int,
                  'completer': orch_ver_completer},
                 {'argument': '--detail', 'action': 'store_true'},
                 [{'argument': '--like'},
                  {'argument': '--id', 'completer': orch_completer},
                  {'argument': '--name', 'completer': orch_name_completer}],
                 orch_list
                 ],
        'create': [{'argument': 'name'},
                   {'argument': '--prompt', 'action': "store_true",
                    'help': 'ask for every orch parameter one by one'},
                   orch_create],
        'copy': [{'argument': 'orchestration_id', 'completer': orch_completer},
                 orch_copy],
        'load': [{'argument': 'file', 'type': argparse.FileType('r')},
                 orch_load],
        'run': [{'argument': 'orchestration_id', 'completer': orch_completer},
                {'argument': '--target', 'metavar': 'TARGET=VALUE', 'action': DictAction, 'nargs': "+", 'dest': 'hosts',
                 'completer': merge_completers([server_completer, granule_completer]),
                 'help': "Run the orch agains the specified target. If no target specified, hosts will be added to "
                         "'all' target. Example: --target node1 node2 backend=node2,node3 "},
                {'argument': '--param', 'metavar': 'PARAM=VALUE', 'action': ParamAction, 'nargs': "+",
                 'dest': 'params', 'help': 'Parameters passed to the orchestration'},
                {'argument': '--no-wait', 'dest': 'background', 'action': 'store_true'},
                orch_run],
    },
    'action': {
        'list': [{'argument': '--version', 'action': 'store', 'type': int,
                  'completer': action_ver_completer},
                 [{'argument': '--json', 'action': 'store_true'},
                  {'argument': '--table', 'action': 'store_true'}],
                 [{'argument': '--like'},
                  {'argument': '--id', 'completer': action_completer},
                  {'argument': '--name', 'completer': action_name_completer}],
                 action_list
                 ],
        'create': [{'argument': 'name'},
                   {'argument': 'action_type', 'choices': [at.name for at in ActionType if at.name != 'NATIVE']},
                   {'argument': '--prompt', 'action': "store_true",
                    'help': 'does not ask for every action parameter one by one'},
                   action_create],
        'copy': [{'argument': 'action_template_id', 'completer': action_completer},
                 action_copy],
        'load': [{'argument': 'file', 'type': argparse.FileType('r')},
                 action_load],
    },
    'exec': {
        'list': [{'argument': '--orch', 'completer': orch_completer},
                 {'argument': '--id', 'dest': 'execution_id'},
                 {'argument': '--server', 'completer': server_completer},
                 {'argument': '--last', 'metavar': 'N', 'type': int, 'help': "shows last N orchestrations"},
                 {'argument': '--asc', 'action': 'store_true'},
                 {'argument': '--detail', 'action': 'store_true'},
                 exec_list]
    },
    'software': {
        'add': [{'argument': 'name'},
                {'argument': 'version'},
                {'argument': 'file'},
                {'argument': '--family', 'completer': software_family_completer},
                software_add],
        'list': [{'argument': '--version', 'action': 'store', 'type': int,
                  'completer': software_ver_completer},
                 {'argument': '--detail', 'action': 'store_true'},
                 [{'argument': '--json', 'action': 'store_true'},
                  {'argument': '--table', 'action': 'store_true'}],
                 [{'argument': '--like'},
                  {'argument': '--id', 'completer': software_completer},
                  {'argument': '--name', 'completer': software_name_completer}],
                 software_list
                 ],
        'send': [{'argument': 'software_id', 'completer': software_completer},
                 {'argument': 'dest_server_id', 'completer': server_completer},
                 {'argument': '--dest_path'},
                 {'argument': '--foreground', 'action': 'store_false', 'dest': 'background', 'default': True},
                 {'argument': '--force', 'action': 'store_true'},
                 software_send], },
    'transfer': {
        'cancel': [{'argument': 'transfer_id'},
                   transfer_cancel],
        'list': [
            {'argument': '--status', 'action': "append", 'nargs': '+', 'choices': [s.name for s in Status]},
            [{'argument': '--id'},
             {'argument': '--last', 'action': 'store', 'type': int}, ],
            transfer_list
        ]},
    'cmd': [{'argument': 'command', 'nargs': '*'},
            {'argument': '--shell', 'action': 'store_true'},
            {'argument': '--target', 'action': ExtendAction, 'nargs': "+",
             'completer': merge_completers([server_completer, granule_completer])},
            {'argument': '--timeout', 'type': int, 'help': 'timeout in seconds to wait for command to terminate'},
            {'argument': '--input'},
            cmd],
    'logfed': {
        'subscribe': {'log': [{'argument': 'src_server_id',
                               'metavar': 'source_server',
                               'help': 'source server to get the logs from',
                               'completer': server_name_completer},
                              {'argument': 'target',
                               'metavar': 'file',
                               'help': 'log file to watch out'},
                              {'argument': 'dst_server_id',
                               'metavar': 'destination_server',
                               'help': 'destination server to get the logs from',
                               'completer': server_name_completer},
                              {'argument': '--mode', 'choices': ['REPO_MIRROR', 'REPO_ROOT', 'MIRROR'],
                                      'help': 'defines where the log will be sent on destination. REPO_MIRROR send the '
                                              'file inside the dest LOG folder and mantains absolute path from origin. '
                                              'REPO_ROOT sends the file inside the dest LOG without mantaining origin'
                                              '\'s path. MIRROR preserves source path and tries to create at dest'},
                              {'argument': '--dest_folder',
                                      'help': 'destination folder to send logs. If not specified, '
                                              'default mode REPO_MIRROR is used'},
                              logfed_subscribe
                              ],
                             'dir': [{'argument': 'target',
                                      'metavar': 'folder',
                                      'help': 'folder to watch out for files to send'},
                                     {'argument': 'src_server_id',
                                      'metavar': 'source_server_id',
                                      'help': 'source server to get the logs from',
                                      'completer': server_completer},
                                     {'argument': 'dst_server_id',
                                      'metavar': 'destination_server_id',
                                      'help': 'source server to get the logs from',
                                      'completer': server_completer},
                                     {'argument': '--include',
                                      'help': 'regular expression used to filter which '
                                              'files and folders should subscribe'},
                                     {'argument': '--exclude',
                                      'help': 'regular expression used to filter which files and folders '
                                              'should NOT subscribe'},
                                     {'argument': '--recursive',
                                      'help': 'enters in each folder to find files to send'},
                                     {'argument': '--mode', 'choices': ['REPO_MIRROR', 'REPO_ROOT', 'MIRROR'],
                                      'help': 'defines where the log will be sent on destination. REPO_MIRROR send the '
                                              'file inside the dest LOG folder and mantains absolute path from origin. '
                                              'REPO_ROOT sends the file inside the dest LOG without mantaining origin'
                                              '\'s path. MIRROR preserves source path and tries to create at dest'},
                                     {'argument': '--dest_folder',
                                      'help': 'destination folder to send logs. If not specified, '
                                              'default mode REPO_MIRROR is used'},
                                     logfed_subscribe]
                      },
        'unsubscribe': [{'argument': 'log_id', 'completer': logfed_completer},
                        logfed_unsubscribe],
        'list': [logfed_list],
    },
    "manager": {
        "catalog": {"refresh": [manager_catalog_refresh]},
        "locker": {"ignore": [{'argument': 'server_ids', 'metavar': 'NODE', 'nargs': '+',
                               'completer': server_completer},
                              functools.partial(manager_locker_ignore, True)],
                   "unignore": [{'argument': 'server_ids', 'metavar': 'NODE', 'nargs': '+',
                                 'completer': server_completer},
                                functools.partial(manager_locker_ignore, False)],
                   'show': [{'argument': 'node', 'nargs': '+', 'completer': server_completer}, manager_locker_show],
                   'unlock': [{'argument': 'scope', 'choices': [s.name for s in Scope]},
                              {'argument': 'node', 'nargs': '+', 'completer': server_completer},
                              manager_locker_unlock]
                   },
        "token": [{'argument': 'expires_time', 'nargs': '?', 'metavar': 'MINUTES',
                   'help': 'Join token expire time in minutes'},
                  {'argument': '--raw', 'action': 'store_true'},
                  manager_token],
        "save": {"login": [manager_save_login]},
        "login": [{'argument': 'username', 'nargs': '?'},
                  manager_login], }
    # "locker": {"lock": [{'argument': "scope", 'choices': ["CATALOG", "ORCHESTRATION", "UPGRADE"]},
    #                     {'argument': "servers", 'nargs': '*', 'completer': server_completer},
    #                     locker_lock],
    #            "unlock": [{'argument': "scope", 'choices': ["CATALOG", "ORCHESTRATION", "UPGRADE"]},
    #                       {'argument': "servers", 'nargs': '*', 'completer': server_completer},
    #                       locker_unlock]},
    ,
    "logging": [{'argument': 'logger', 'completer': logger_completer},
                {'argument': 'level', 'choices': ['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL']},
                logging_cmd],
    "env": {
        "list": [env_list],
        "get": [{'argument': 'key', 'completer': DshellWordCompleter(environ._environ.keys())},
                env_get],
        "set": [{'argument': 'key', 'completer': DshellWordCompleter(environ._environ.keys())},
                {'argument': 'value', 'nargs': '*'},
                env_set]},
    'exit': [exit_dshell]

}



import concurrent
import os
import shlex

from prompt_toolkit import prompt
from prompt_toolkit.auto_suggest import AutoSuggestFromHistory
from prompt_toolkit.history import InMemoryHistory

from dimensigon.dshell import network as ntwrk
from dimensigon.dshell.argparse_raise import ArgumentParserRaise
from dimensigon.dshell.completer import server_completer
from dimensigon.dshell.helpers import get_history
from dimensigon.dshell.output import dprint
from dimensigon.dshell.prompts.utils import prompt_parameter
from dimensigon.dshell.validators import IntValidator, ListConverter

form = {
    "target": dict(history=InMemoryHistory(), completer=server_completer, validator=ListConverter()),
    "timeout": dict(history=InMemoryHistory(), validator=IntValidator()),
    "input": dict(history=InMemoryHistory()),
}

parser = ArgumentParserRaise(allow_abbrev=False, prog='')
subparser = parser.add_subparsers(dest='cmd')
target_parser = subparser.add_parser('target')
target_parser.add_argument('servers')
run_parser = subparser.add_parser('run')
exit_action = subparser.add_parser('exit')

history = None

entity_name = os.path.basename(__file__).rstrip('.py')

ids = []
name2id = {}


def fill_data():
    global ids, name2id
    resp = ntwrk.get('api_1_0.serverlist', timeout=15)
    if resp.ok:
        ids = []
        name2id = {}
        for s in resp.msg:
            name2id[s.get('name')] = s.get('id')
            ids.append(s.get('id'))
    else:
        return resp


def execute_cmd(command: str, target, timeout=None, input=None):
    global name2id, ids
    data = {'command': command, 'target': target}
    if timeout:
        data.update(timeout=timeout)
    if input:
        data.update(input=input.replace('\\n', '\n').replace('\\t', '\t'))
    if target in name2id:
        dest = name2id[target]
    elif target in ids:
        dest = target
    else:
        raise ValueError(f'invalid target {target}')
    resp = ntwrk.post('api_1_0.launch_command', view_data={'params': 'human'}, json=data,  headers={'D-Destination': dest})
    return resp


def subprompt(entity, changed=False, ask_all=False, parent_prompt=None):
    global history, ids, name2id

    with concurrent.futures.ThreadPoolExecutor(max_workers=6) as executor:
        if history is None:
            history = get_history(entity_name, InMemoryHistory())

        if ask_all:
            for k, v in form.items():
                try:
                    if prompt_parameter(k, entity, form,
                                        parent_prompt=f"{parent_prompt} {entity_name}"):
                        changed = True
                except KeyboardInterrupt:
                    entity[k] = None
                except EOFError:
                    break

        while True:
            try:
                text = prompt(f"{parent_prompt} {entity_name}> ", history=history, auto_suggest=AutoSuggestFromHistory())
            except KeyboardInterrupt:
                continue
            except EOFError:
                return

            splitted = shlex.split(text)
            if not splitted:
                continue
            if splitted[0] == 'target':
                if splitted[1:]:
                    resp = fill_data()
                    if resp:
                        dprint(resp)
                        continue
                    unknown = [s for s in splitted[1:] if s not in name2id and s not in ids]
                    if unknown:
                        dprint(f"Unknown target {', '.join(unknown)}")
                    else:
                        entity['target'] = splitted[1:]
                else:
                    dprint(' '.join(entity['target']))
            elif splitted[0] == 'set':
                if len(splitted[1:]) == 1 and splitted[1] in form:
                    try:
                        if prompt_parameter(splitted[1], entity, form,
                                            f"{parent_prompt}{entity_name}"):
                            changed = True
                    except KeyboardInterrupt:
                        continue  # Control-C pressed. Try again.
                    except EOFError:
                        return
                else:
                    dprint("Not a valid parameter. Available: "+', '.join(form.keys()))

            elif splitted[0] == 'exit':
                return
            else:
                resp = fill_data()
                if resp:
                    dprint(resp)
                    continue
                if entity['target']:
                    first = True
                    future_to_server = {
                        executor.submit(execute_cmd, text, server, timeout=entity.get('timeout'),
                                        input=entity.get('input')): server for server in
                        entity['target']}
                    for future in concurrent.futures.as_completed(future_to_server):
                        server = future_to_server[future]
                        try:
                            resp = future.result()
                        except Exception as exc:
                            dprint(exc)
                        else:
                            if first:
                                if "cmd" in resp.msg:
                                    resp.msg.pop('cmd')
                                if "input" in resp.msg:
                                    input = resp.msg.pop('input')
                                    if input:
                                        dprint(f"input: {input}")
                                first = False
                            else:
                                resp.msg.pop('cmd', None)
                                resp.msg.pop('input', None)
                            dprint(resp)
                else:
                    dprint("no target specified")



import copy
import datetime as dt
import random
import threading
import typing as t

from dataclasses import dataclass

from dimensigon.utils.helpers import get_now
from dimensigon.utils.typos import Id

date_format = "%m/%d/%Y, %H:%M:%S %z"

@dataclass
class _ClusterRegister:
    id: Id
    birth: dt.datetime = None
    keepalive: dt.datetime = None
    death: dt.datetime = None
    coordinator: bool = False

    def to_dict(self, dateformat=date_format):
        data = {'id': self.id}
        if self.birth:
            data.update(birth=self.birth.strftime(dateformat))
        if self.keepalive:
            data.update(keepalive=self.keepalive.strftime(dateformat))
        if self.death:
            data.update(death=self.death.strftime(dateformat))
        if self.coordinator:
            data.update(coordinator=True)
        return data

    def __eq__(self, other: '_ClusterRegister'):
        if isinstance(other, self.__class__):
            return self.id == other.id and self.birth == other.birth \
                   and self.keepalive == other.keepalive \
                   and self.death == other.death \
                   and self.coordinator == other.coordinator
        else:
            return False

    @staticmethod
    def from_dict(dto, dateformat=date_format):
        ident = dto['id']
        if dto.get('birth', None):
            birth = dt.datetime.strptime(dto['birth'], dateformat)
        else:
            birth = None

        if dto.get('keepalive', None):
            keepalive = dt.datetime.strptime(dto['keepalive'], dateformat)
        else:
            keepalive = None

        if dto.get('death', None):
            death = dt.datetime.strptime(dto['death'], dateformat)
        else:
            death = None

        return _ClusterRegister(ident, birth=birth, keepalive=keepalive, death=death)


@dataclass
class _ClusterRegisterSession(_ClusterRegister):
    session: Id = None

    def to_dict(self, dateformat=date_format):
        data = super().to_dict(dateformat)
        data['session'] = self.session
        return data

    def __eq__(self, other: '_ClusterRegisterSession'):
        return super().__eq__(other) and self.session == other.session

    @classmethod
    def from_dict(cls, dto, dateformat=date_format):
        kwargs = dict(id=dto['id'], session=dto['session'])
        if dto.get('birth', None):
            kwargs['birth'] = dt.datetime.strptime(dto['birth'], dateformat)

        if dto.get('keepalive', None):
            kwargs['keepalive'] = dt.datetime.strptime(dto['keepalive'], dateformat)

        if dto.get('death', None):
            kwargs['death'] = dt.datetime.strptime(dto['death'], dateformat)

        return cls(**kwargs)


class Cluster(object):
    _register_class = _ClusterRegister
    _cluster: t.Dict[Id, type(_register_class)] = {}
    _lock = threading.RLock()

    def __init__(self, threshold=None):
        self.threshold = threshold or dt.timedelta()

    def __getitem__(self, iden):
        return self._cluster[iden].to_dict()

    def __contains__(self, iden):
        try:
            cr = self._cluster[iden]
        except KeyError:
            return False
        if cr.death is None:
            return True

    def __iter__(self):
        with self._lock:
            for cr in self._cluster.values():
                if cr.death is None:
                    yield cr.id

    def set_alive(self, iden: Id, alive=None) -> t.Optional[t.Dict]:
        with self._lock:
            alive = alive or get_now()
            initial_cr = copy.deepcopy(self._cluster.get(iden, None))
            if iden in self._cluster:
                if self._cluster[iden].death is not None:
                    if self._cluster[iden].birth < alive:
                        self._cluster[iden].birth = alive
                    self._cluster[iden].keepalive = alive
                    self._cluster[iden].death = None
                elif alive < self._cluster[iden].birth:
                    self._cluster[iden].birth = alive
                else:
                    return None
            else:
                self._cluster[iden] = self._register_class(iden, birth=alive, keepalive=alive)
            return self._cluster[iden].to_dict() if initial_cr != self._cluster[iden] else None

    def set_keepalive(self, iden: Id, keepalive=None) -> t.Optional[t.Dict]:
        with self._lock:
            keepalive = keepalive or get_now()
            initial_cr = copy.deepcopy(self._cluster.get(iden, None))
            if iden in self._cluster:
                if self._cluster[iden].death is not None:
                    self.set_alive(iden)
                if not self._cluster[iden].keepalive or self._cluster[iden].keepalive < keepalive - self.threshold:
                    self._cluster[iden].keepalive = keepalive
            else:
                self._cluster[iden] = self._register_class(iden, birth=keepalive, keepalive=keepalive)
            return self._cluster[iden].to_dict() if self._cluster[iden] != initial_cr else None

    def set_death(self, iden: Id, death=None) -> t.Optional[t.Dict]:
        with self._lock:
            death = death or get_now()
            if iden in self._cluster:
                cr = self._cluster[iden]
                if cr.death is None:
                    if death > cr.birth:
                        cr.death = death
                    else:
                        cr.death = cr.birth
                else:
                    if death and cr.death > death:
                        cr.death = death
                    else:
                        return None
            else:
                self._cluster[iden] = self._register_class(iden, birth=death, death=death)
            return self._cluster[iden].to_dict()

    def update_cluster(self, in_cr_list: t.Union[dict, list]) -> t.List[_ClusterRegister]:
        if isinstance(in_cr_list, list):
            return self._update_cluster([self._register_class.from_dict(cr) for cr in in_cr_list])
        else:
            return self._update_cluster([self._register_class.from_dict(in_cr_list)])

    def _update_cluster(self, in_cr_list: t.List[_ClusterRegister]):
        with self._lock:
            changed = []
            for in_cr in in_cr_list:
                updated = False
                if in_cr.id in self._cluster:
                    cr = self._cluster[in_cr.id]
                    if in_cr.birth > cr.birth:
                        self._cluster[in_cr.id] = in_cr
                        cr = self._cluster[in_cr.id]
                        updated = True
                    if in_cr.keepalive is not None and (cr.keepalive is None or in_cr.keepalive > cr.keepalive):
                        cr.keepalive = in_cr.keepalive
                        cr.death = None
                        updated = True
                    if in_cr.death is not None and cr.death is not None and in_cr.death > cr.death:
                        cr.death = in_cr.death
                        updated = True
                    if cr.death is None and in_cr.death is not None and in_cr.death > cr.birth:
                        cr.death = in_cr.death
                        updated = True
                    elif in_cr.birth is not None and cr.death is not None and in_cr.birth > cr.death:
                        cr.birth = in_cr.birth
                        cr.death = in_cr.death
                        updated = True
                else:
                    self._cluster[in_cr.id] = in_cr
                    updated = True
                if updated:
                    changed.append(in_cr.id)
            return changed

    def get_alive(self) -> t.List[Id]:
        with self._lock:
            return [cr.id for cr in self._cluster.values() if cr.birth is not None and cr.death is None]

    def get_delta_keepalive(self, delta: dt.timedelta):
        with self._lock:
            now = get_now()
            return [cr.id for cr in self._cluster.values() if
                    cr.birth is not None and cr.death is None and (now - (cr.keepalive or cr.birth)) < delta]

    def get_oldest_alive(self) -> t.Union[Id, None]:
        with self._lock:
            cr_list = [cr for cr in self._cluster.values() if cr.death is None]
            cr_list.sort(key=lambda x: x.birth)
            return cr_list[0].id

    def get_cluster(self) -> t.List[t.Dict]:
        with self._lock:
            return [cr.to_dict() for cr in self._cluster.values()]

    def get(self, iden) -> t.Dict:
        with self._lock:
            return self._cluster[iden].to_dict()

    # only for testing purposes
    def clear_cluster(self):
        self._cluster = {}


class ClusterKeepAlive(Cluster):

    def set_alive(self, iden: Id, alive=None) -> t.Optional[t.Dict]:
        with self._lock:
            alive = alive or get_now()
            if iden in self._cluster:
                if self._cluster[iden].death is not None:
                    self._cluster[iden].birth = alive
                    self._cluster[iden].death = None
                else:
                    self._cluster[iden].birth = alive
            else:
                self._cluster[iden] = self._register_class(iden, birth=alive)
            return self._cluster[iden].to_dict()


class ClusterCoordinator(Cluster):
    _cluster: t.Dict[Id, _ClusterRegisterSession] = {}
    _register_class = _ClusterRegister

    def set_alive(self, iden: Id) -> t.Dict:
        with self._lock:
            super().set_alive(iden)
            if not self.coordinators:
                self._cluster[iden].coordinator = True
            return self._cluster[iden].to_dict()

    def set_death(self, iden: Id, death=None) -> t.List[t.Dict]:
        with self._lock:
            death = super().set_death(iden, death)
            data = [death]
            if self._cluster[iden].coordinator:
                self._cluster[iden].coordinator = False
                death['coordinator'] = False
                new_coordinator = self.try_set_coordinator(self.get_oldest_alive())
                if new_coordinator:
                    data.append(self._cluster[self.get_oldest_coordinator()].to_dict())
            return data


    @property
    def coordinators(self) -> t.List[_ClusterRegister]:
        with self._lock:
            return [cr for cr in self._cluster.values() if cr.coordinator and cr.death is None]

    def get_coordinators(self) -> t.List[Id]:
        with self._lock:
            return [cr.id for cr in self._cluster.values() if cr.coordinator and cr.death is None]

    def get_oldest_coordinator(self) -> t.Union[Id, None]:
        with self._lock:
            coord_cr_list = [cr for cr in self._cluster.values() if cr.coordinator and cr.death is None]
        if coord_cr_list:
            coord_cr_list.sort(key=lambda x: x.birth)
            return coord_cr_list[0].id
        else:
            return None

    def try_set_coordinator(self, iden: Id):
        with self._lock:
            if not self.coordinators:
                self._cluster[iden].coordinator = True
                return True
            else:
                return False

    def set_coordinator(self, iden: Id):
        with self._lock:
            updated = False
            for cr in self._cluster.values():
                if cr.coordinator:
                    updated = True
                    cr.coordinator = False
            if not self._cluster[iden].coordinator:
                self._cluster[iden].coordinator = True
                updated = True
            return updated

    def _update_cluster(self, in_cr_list: t.List[_ClusterRegister]):
        with self._lock:
            updated = super()._update_cluster(in_cr_list)
            for in_cr in in_cr_list:
                if self._cluster[in_cr.id].death is None:
                    self._cluster[in_cr.id].coordinator = self._cluster[in_cr.id].coordinator or in_cr.coordinator
                    if in_cr.coordinator and not self._cluster[in_cr.id].coordinator:
                        updated = True

            # select oldest coordinators
            coordinator = self.get_oldest_coordinator()
            if coordinator is None:
                cr_l = list(self._cluster.values())
                cr_l.sort(key=lambda x: x.birth)
                if len(cr_l) > 0:
                    coordinator = cr_l[0].id
            if coordinator:
                coord_updated = self.set_coordinator(coordinator)

            return updated or coord_updated


class ClusterSession(Cluster):
    _register_class = _ClusterRegisterSession

    def __init__(self, threshold: dt.timedelta = None):
        self._session = None
        self.threshold = threshold

    def set_session(self):
        if self._session is None:
            self._session = int(random.random() * 1000000000000)

    @property
    def session(self):
        return self._session

    def set_alive(self, iden: Id, session: Id, alive=None) -> t.Optional[t.Dict]:
        with self._lock:
            alive = alive or get_now()
            if iden in self._cluster:
                if self._cluster[iden].session == session:
                    if self._cluster[iden].death is not None:
                        if alive < self._cluster[iden].birth:
                            self._cluster[iden].birth = alive
                        self._cluster[iden].keepalive = alive
                        self._cluster[iden].death = None
                    else:
                        return None
                else:
                    self._cluster[iden].session = session
                    self._cluster[iden].birth = alive
                    self._cluster[iden].keepalive = alive
                    self._cluster[iden].death = None
            else:
                self._cluster[iden] = self._register_class(iden, session=session, birth=alive, keepalive=alive)
            return self._cluster[iden].to_dict()

    def set_death(self, iden: Id, session: Id, death=None) -> t.Optional[t.Dict]:
        with self._lock:
            death = death or get_now()
            initial_cr = copy.deepcopy(self._cluster.get(iden, None))
            if iden in self._cluster:
                cr = self._cluster[iden]
                if self._cluster[iden].session == session:
                    if cr.death is None:
                        if death > cr.birth:
                            cr.death = death
                        else:
                            cr.death = cr.birth
                    else:
                        if death and cr.death > death:
                            cr.death = death
                        else:
                            return None
                else:

                    if cr.death is None:
                        if cr.birth < death:
                            self._cluster[iden].death = death
                            self._cluster[iden].birth = death
                            self._cluster[iden].session = session
                    else:
                        if cr.death < death:
                            self._cluster[iden].death = death
                            self._cluster[iden].birth = death
                            self._cluster[iden].session = session
            else:
                self._cluster[iden] = self._register_class(iden, session=session, birth=death, death=death)
            if initial_cr != self._cluster[iden]:
                return self._cluster[iden].to_dict()

    def set_keepalive(self, iden: Id, session: Id = None, keepalive=None) -> t.Optional[t.Dict]:
        with self._lock:
            keepalive = keepalive or get_now()
            if iden in self._cluster:
                if self._cluster[iden].session == session:
                    self.set_alive(iden, session, keepalive)
                self._cluster[iden].keepalive = keepalive
            else:
                self._cluster[iden] = self._register_class(iden, session=session, birth=keepalive, keepalive=keepalive)
            return self._cluster[iden].to_dict()

    def _update_cluster(self, in_cr_list: t.List[_ClusterRegisterSession]):
        with self._lock:
            updated = False
            for in_cr in in_cr_list:
                if in_cr.id in self._cluster:
                    cr = self._cluster[in_cr.id]
                    if cr.session == in_cr.session:
                        if in_cr.birth < cr.birth:
                            cr.birth = in_cr.birth
                            updated = True
                        if in_cr.keepalive is not None and (cr.keepalive is None or in_cr.keepalive > cr.keepalive):
                            cr.keepalive = in_cr.keepalive
                            updated = True
                        if in_cr.death is not None and cr.death is not None and in_cr.death < cr.death:
                            cr.death = in_cr.death
                            updated = True
                        if cr.death is None and in_cr.death is not None and in_cr.death > cr.birth:
                            cr.death = in_cr.death
                            updated = True
                        elif in_cr.birth is not None and cr.death is not None and in_cr.birth > cr.death:
                            cr.birth = in_cr.birth
                            cr.death = in_cr.death
                            updated = True
                    else:
                        # newer session is stored
                        if cr.birth < in_cr.birth:
                            self._cluster[in_cr.id] = in_cr
                            updated = True
                else:
                    self._cluster[in_cr.id] = in_cr
                    updated = True
            return updated





from dimensigon.utils.helpers import get_now
from dimensigon.utils.typos import UtcDateTime
from dimensigon.web import db


class SchemaChanges(db.Model):
    """Representation of schema version changes."""

    __tablename__ = "L_schema_changes"
    change_id = db.Column(db.Integer, primary_key=True)
    schema_version = db.Column(db.Integer)
    changed = db.Column(UtcDateTime(timezone=True), default=get_now)

    def set_initial(self):
        from dimensigon.domain.entities import SCHEMA_VERSION
        count = self.query.count()
        # database started
        if count == 0:
            sc = SchemaChanges(schema_version=SCHEMA_VERSION)
            db.session.add(sc)



import copy
import datetime as dt
import ipaddress
import logging
import socket
import typing as t

from flask import current_app, url_for, g, has_app_context
from sqlalchemy import or_
from sqlalchemy.orm.exc import NoResultFound

from dimensigon.utils.typos import ScalarListType, Gate as TGate, UtcDateTime, Id
from dimensigon.web import db, errors
from .base import UUIDistributedEntityMixin, SoftDeleteMixin
from .gate import Gate
from .route import Route, RouteContainer
from ... import defaults
from ...utils.helpers import get_ips, get_now, is_iterable_not_string
from ...web.helpers import QueryWithSoftDelete

_me = {}

class Server(db.Model, UUIDistributedEntityMixin, SoftDeleteMixin):
    __tablename__ = 'D_server'
    order = 10

    name = db.Column(db.String(255), nullable=False, unique=True)
    granules = db.Column(ScalarListType())
    _me = db.Column("me", db.Boolean, default=False)
    _old_name = db.Column("$$name", db.String(255))
    l_ignore_on_lock = db.Column("ignore_on_lock", db.Boolean,
                                 default=False)  # ignore the server for locking when set
    created_on = db.Column(UtcDateTime(timezone=True))  # new in version 3

    route: t.Optional[Route] = db.relationship("Route", primaryjoin="Route.destination_id==Server.id", uselist=False,
                                               back_populates="destination", cascade="all, delete-orphan")
    gates: t.List[Gate] = db.relationship("Gate", back_populates="server", cascade="all, delete-orphan")

    # software_list = db.relationship("SoftwareServerAssociation", back_populates="server")

    query_class = QueryWithSoftDelete

    def __init__(self, name: str, granules: t.List[str] = None,
                 dns_or_ip: t.Union[str, ipaddress.IPv4Address, ipaddress.IPv6Address] = None, port: int = None,
                 gates: t.List[t.Union[TGate, t.Dict[str, t.Any]]] = None, me: bool = False, created_on=None,
                 **kwargs):
        UUIDistributedEntityMixin.__init__(self, **kwargs)
        SoftDeleteMixin.__init__(self, **kwargs)

        self.name = name
        if port or dns_or_ip:
            self.add_new_gate(dns_or_ip or self.name, port or defaults.DEFAULT_PORT)

        if gates:
            for gate in gates:
                if isinstance(gate, str):
                    if ':' in gate:
                        self.add_new_gate(*gate.split(':'))
                    else:
                        self.add_new_gate(gate, defaults.DEFAULT_PORT)
                elif isinstance(gate, tuple):
                    self.add_new_gate(*gate)
                elif isinstance(gate, dict):
                    gate['server'] = self
                    Gate.from_json(gate)

        assert 'all' not in (granules or [])
        self.granules = granules or []
        self._me = me
        self.created_on = created_on or get_now()
        # create an empty route
        if not me and not self.deleted:
            Route(self)


    @property
    def external_gates(self):
        e_g = []
        for g in self.gates:
            if not g.ip:
                try:
                    ip = ipaddress.ip_address(
                        socket.getaddrinfo(g.dns, 0, family=socket.AF_INET, proto=socket.IPPROTO_TCP)[0][4][0])
                except socket.gaierror:
                    e_g.append(g)
                    continue
                except KeyError:
                    e_g.append(g)
                    continue
            else:
                ip = g.ip
            if not ip.is_loopback:
                e_g.append(g)
        return e_g

    @property
    def localhost_gates(self):
        l_g = []
        for g in self.gates:
            if not g.ip:
                try:
                    ip = ipaddress.ip_address(
                        socket.getaddrinfo(g.dns, 0, family=socket.AF_INET, proto=socket.IPPROTO_TCP)[0][4][0])
                except socket.gaierror:
                    continue
                except KeyError:
                    continue
            else:
                ip = g.ip
            if ip.is_loopback:
                l_g.append(g)
        return l_g

    @property
    def hidden_gates(self):
        hg = [g for g in self.gates if g.hidden]
        hg.sort(key=lambda x: x.last_modified_at or get_now())
        return hg

    def add_new_gate(self, dns_or_ip: t.Union[str, ipaddress.IPv4Address, ipaddress.IPv6Address], port: int,
                     hidden=None):
        ip = None
        dns = None
        if dns_or_ip:
            try:
                ip = ipaddress.ip_address(dns_or_ip)
            except ValueError:
                dns = dns_or_ip
        return Gate(server=self, port=port, dns=dns, ip=ip, hidden=hidden)

    def __str__(self, ):
        return f"{self.name}"

    def url(self, view: str = None, **values) -> str:
        """
        generates the full url to access the server. Uses url_for to generate the full_path.

        Parameters
        ----------
        view
        values

        Raises
        -------
        ConnectionError:
            if server is unreachable
        """
        scheme = 'http' if current_app.dm and 'keyfile' not in current_app.dm.config.http_conf else 'https'
        gate = None
        route = self.route
        if self._me and (route is None or route.cost is None):
            try:
                gate = self.localhost_gates[0]
            except IndexError:
                current_app.logger.warning(
                    f"No localhost set for '{self}'. Trying connection through another gate")
                if len(self.gates) == 0:
                    raise RuntimeError(f"No gate set for server '{self}'")
                gate = self.gates[0]
        elif getattr(route, 'cost', None) == 0:
            gate = route.gate
        elif getattr(route, 'proxy_server', None):
            gate = getattr(getattr(route.proxy_server, 'route', None), 'gate', None)

        if not gate:
            raise errors.UnreachableDestination(self, getattr(g, 'server', None))

        root_path = f"{scheme}://{gate}"

        if view is None:
            return root_path
        else:
            with current_app.test_request_context():
                return root_path + url_for(view, **values)

    @classmethod
    def get_neighbours(cls, alive=False,
                       exclude: t.Union[t.Union[Id, 'Server'], t.List[t.Union[Id, 'Server']]] = None) -> t.List[
        'Server']:
        """returns neighbour servers

        Args:
            alive: if True, returns neighbour servers inside the cluster

        Returns:

        """
        query = cls.query.join(cls.route).filter(Route.cost == 0)
        if exclude:
            if isinstance(exclude, list):
                if isinstance(exclude[0], Server):
                    query = query.filter(Server.id.in_([s.id for s in exclude]))
                else:
                    query = query.filter(Server.id.in_(exclude))
            elif isinstance(exclude, Server):
                query = query.filter(Server.id != exclude.id)
            else:
                query = query.filter(Server.id != exclude)

        if alive:
            query = query.filter(Server.id.in_([iden for iden in current_app.cluster_manager.cluster.get_delta_keepalive(
                dt.timedelta(minutes=defaults.COMA_NODE_FACTOR * defaults.REFRESH_PERIOD))]))

        return query.order_by(Server.name).all()

    @classmethod
    def get_neighbours(cls, alive=False,
                       exclude: t.Union[t.Union[Id, 'Server'], t.List[t.Union[Id, 'Server']]] = None) -> t.List[
        'Server']:
        """returns neighbour servers

        Args:
            alive: if True, returns neighbour servers inside the cluster

        Returns:

        """
        query = cls.query.join(cls.route).filter(Route.cost == 0)
        if exclude:
            if isinstance(exclude, list):
                if isinstance(exclude[0], Server):
                    query = query.filter(Server.id.in_([s.id for s in exclude]))
                else:
                    query = query.filter(Server.id.in_(exclude))
            elif isinstance(exclude, Server):
                query = query.filter(Server.id != exclude.id)
            else:
                query = query.filter(Server.id != exclude)

        if alive:
            query = query.filter(Server.id.in_([iden for iden in current_app.cluster.cluster_manager.get_delta_keepalive(
                dt.timedelta(minutes=defaults.COMA_NODE_FACTOR * defaults.REFRESH_PERIOD))]))

        return query.order_by(Server.name).all()
    @classmethod
    def get_not_neighbours(cls) -> t.List['Server']:
        return cls.query.outerjoin(cls.route).filter(
            or_(or_(Route.cost > 0, Route.cost == None), cls.route == None)).filter(Server._me == False).order_by(
            Server.name).all()

    @classmethod
    def get_reachable_servers(cls, alive=False, exclude: t.Union[t.Union[Id, 'Server'], t.List[t.Union[Id, 'Server']]] = None) -> t.List[
        'Server']:
        """returns list of reachable servers

        Args:
            alive: if True, returns servers inside the cluster
            exclude: filter to exclude servers

        Returns:
        list of all reachable servers
        """
        query = cls.query.join(cls.route).filter(Route.cost.isnot(None))
        if exclude:
            if is_iterable_not_string(exclude):
                c_exclude = [e.id if isinstance(e, Server) else e for e in exclude]
            else:
                c_exclude = [exclude.id if isinstance(exclude, Server) else exclude]
            query = query.filter(Server.id.notin_(c_exclude))

        if alive:
            query = query.filter(Server.id.in_([iden for iden in current_app.cluster_manager.cluster.get_delta_keepalive(
                dt.timedelta(minutes=defaults.COMA_NODE_FACTOR * defaults.REFRESH_PERIOD))]))

        return query.order_by(Server.name).all()

    def to_json(self, add_gates=False, human=False, add_ignore=False):
        data = super().to_json()
        data.update(
            {'name': self.name, 'granules': self.granules,
             'created_on': self.created_on.strftime(defaults.DATETIME_FORMAT)})
        if add_gates:
            data.update(gates=[])
            for g in self.gates:
                json_gate = g.to_json(human=human)
                json_gate.pop('server_id', None)
                json_gate.pop('server', None)  # added to remove when human set
                data['gates'].append(json_gate)
        if add_ignore:
            data.update(ignore_on_lock=self.l_ignore_on_lock)
        return data

    @classmethod
    def from_json(cls, kwargs) -> 'Server':
        kwargs = copy.deepcopy(kwargs)
        gates = kwargs.pop('gates', [])
        if 'created_on' in kwargs:
            kwargs['created_on'] = dt.datetime.strptime(kwargs['created_on'], defaults.DATETIME_FORMAT)
        server = super().from_json(kwargs)
        for gate in gates:
            gate.update(server=server)
            Gate.from_json(gate)
        return server

    @classmethod
    def get_current(cls) -> 'Server':
        global _me
        if has_app_context():
            app = current_app._get_current_object()
            if app not in _me:
                entity = cls.query.filter_by(_me=True).filter_by(deleted=False).one()
                if entity:
                    db.session.expunge(entity)
                    _me[app] = entity
                else:
                    raise NoResultFound('No row was found for one()')
            return db.session.merge(_me[app], load=False)
        return cls.query.filter_by(_me=True).filter_by(deleted=False).one()

    @staticmethod
    def set_initial(session=None, gates=None):
        logger = logging.getLogger('dimensigon.db')
        if session is None:
            session = db.session
        server = session.query(Server).filter_by(_me=True).all()
        if len(server) == 0:
            try:
                server_name = current_app.config.get('SERVER_NAME') or defaults.HOSTNAME
            except:
                server_name = defaults.HOSTNAME

            if gates is None:
                gates = [(ip, defaults.DEFAULT_PORT) for ip in get_ips()]
            server = Server(name=server_name,
                            gates=gates,
                            me=True)
            logger.info(f'Creating Server {server.name} with the following gates: {gates}')
            session.add(server)
        elif len(server) > 1:
            raise ValueError('Multiple servers found as me.')

    def ignore_on_lock(self, value: bool):
        if value != self.l_ignore_on_lock:
            from dimensigon.domain.entities import bypass_datamark_update
            with bypass_datamark_update:
                self.l_ignore_on_lock = value

    def set_route(self, proxy_route, gate=None, cost=None):
        if self.route is None:
            self.route = Route(destination=self)
        if isinstance(proxy_route, RouteContainer):
            self.route.set_route(proxy_route)
        elif isinstance(proxy_route, Route):
            assert proxy_route.destination == self
            self.route.set_route(RouteContainer(proxy_route.proxy_server, proxy_route.gate, proxy_route.cost))
        else:
            self.route.set_route(RouteContainer(proxy_route, gate, cost))

    def delete(self):
        super().delete()
        for g in self.gates:
            g.delete()


import json
import typing as t
from datetime import datetime

from flask import current_app
from flask_jwt_extended import get_jwt_identity

from dimensigon import defaults
from dimensigon.domain.entities.base import EntityReprMixin, UUIDEntityMixin
from dimensigon.utils.typos import UUID, UtcDateTime
from dimensigon.web import db
from .server import Server
from ...utils.helpers import is_iterable_not_string, get_now

if t.TYPE_CHECKING:
    from dimensigon.use_cases.operations import CompletedProcess


class StepExecution(db.Model, UUIDEntityMixin, EntityReprMixin):
    __tablename__ = 'L_step_execution'

    start_time = db.Column(UtcDateTime(timezone=True), nullable=False)
    end_time = db.Column(UtcDateTime(timezone=True))
    params = db.Column(db.JSON)
    rc = db.Column(db.Integer)
    stdout = db.Column(db.Text)
    stderr = db.Column(db.Text)
    success = db.Column(db.Boolean)
    step_id = db.Column(UUID, db.ForeignKey('D_step.id'), nullable=False)
    server_id = db.Column(UUID, db.ForeignKey('D_server.id'))
    orch_execution_id = db.Column(UUID, db.ForeignKey('L_orch_execution.id'))
    pre_process_elapsed_time = db.Column(db.Float)
    execution_elapsed_time = db.Column(db.Float)
    post_process_elapsed_time = db.Column(db.Float)
    child_orch_execution_id = db.Column(UUID)

    step = db.relationship("Step")
    server = db.relationship("Server", foreign_keys=[server_id])
    orch_execution = db.relationship("OrchExecution", foreign_keys=[orch_execution_id],
                                     uselist=False, back_populates="step_executions")
    child_orch_execution = db.relationship("OrchExecution", uselist=False, foreign_keys=[child_orch_execution_id],
                                           primaryjoin="StepExecution.child_orch_execution_id==OrchExecution.id")

    # def __init__(self, *args, **kwargs):
    #     UUIDEntityMixin.__init__(self, **kwargs)

    def load_completed_result(self, cp: 'CompletedProcess'):
        self.success = cp.success
        self.stdout = cp.stdout
        self.stderr = cp.stderr
        self.rc = cp.rc

    def to_json(self, human=False, split_lines=False):
        data = {}
        if self.id:
            data.update(id=str(self.id))
        if self.start_time:
            data.update(start_time=self.start_time.strftime(defaults.DATETIME_FORMAT))
        if self.end_time:
            data.update(end_time=self.end_time.strftime(defaults.DATETIME_FORMAT))
        data.update(params=self.params)
        data.update(step_id=str(getattr(self.step, 'id', None)) if getattr(self.step, 'id', None) else None)
        data.update(rc=self.rc, success=self.success)
        if human:
            data.update(server=str(self.server) if self.server else None)
            try:
                stdout = json.loads(self.stdout)
            except:
                stdout = self.stdout.split('\n') if split_lines and self.stdout else self.stdout
            try:
                stderr = json.loads(self.stderr)
            except:
                stderr = self.stderr.split('\n') if split_lines and self.stderr else self.stderr
        else:
            data.update(
                server_id=str(getattr(self.server, 'id', None)) if getattr(self.server, 'id', None) else None)
            stdout = self.stdout.split('\n') if split_lines and self.stdout else self.stdout
            stderr = self.stderr.split('\n') if split_lines and self.stderr else self.stderr
        data.update(stdout=stdout)
        data.update(stderr=stderr)
        if self.child_orch_execution_id:
            data.update(child_orch_execution_id=str(self.child_orch_execution_id))
        data.update(
            pre_process_elapsed_time=self.pre_process_elapsed_time) if self.pre_process_elapsed_time is not None else None
        data.update(
            execution_elapsed_time=self.execution_elapsed_time) if self.execution_elapsed_time is not None else None
        data.update(
            post_process_elapsed_time=self.post_process_elapsed_time) if self.post_process_elapsed_time is not None else None
        return data


class OrchExecution(db.Model, UUIDEntityMixin, EntityReprMixin):
    __tablename__ = 'L_orch_execution'

    start_time = db.Column(UtcDateTime(timezone=True), nullable=False, default=get_now)
    end_time = db.Column(UtcDateTime(timezone=True))
    orchestration_id = db.Column(UUID, db.ForeignKey('D_orchestration.id'), nullable=False)
    target = db.Column(db.JSON)
    params = db.Column(db.JSON)
    executor_id = db.Column(UUID, db.ForeignKey('D_user.id'))
    service_id = db.Column(UUID, db.ForeignKey('D_service.id'))
    success = db.Column(db.Boolean)
    undo_success = db.Column(db.Boolean)
    message = db.Column(db.Text)
    server_id = db.Column(UUID, db.ForeignKey('D_server.id'))
    parent_step_execution_id = db.Column(UUID)

    orchestration = db.relationship("Orchestration")
    executor = db.relationship("User")
    service = db.relationship("Service")
    step_executions = db.relationship("StepExecution", back_populates="orch_execution", order_by="StepExecution.start_time")

    server = db.relationship("Server", foreign_keys=[server_id])
    parent_step_execution = db.relationship("StepExecution", uselist=False, foreign_keys=[parent_step_execution_id],
                                            primaryjoin="OrchExecution.parent_step_execution_id==StepExecution.id")

    # def __init__(self, *args, **kwargs):
    #     UUIDEntityMixin.__init__(self, **kwargs)

    def to_json(self, add_step_exec=False, human=False, split_lines=False):
        data = {}
        if self.id:
            data.update(id=str(self.id))
        if self.start_time:
            data.update(start_time=self.start_time.strftime(defaults.DATETIME_FORMAT))
        if self.end_time:
            data.update(end_time=self.end_time.strftime(defaults.DATETIME_FORMAT))
        if human:
            # convert target ids to server names
            d = {}
            if isinstance(self.target, dict):
                for k, v in self.target.items():
                    if is_iterable_not_string(v):
                        d[k] = [str(Server.query.get(s) or s) for s in v]
                    else:
                        d[k] = str(Server.query.get(v) or v)
            elif isinstance(self.target, list):
                d = [str(Server.query.get(s) or s) for s in self.target]
            else:
                d = str(Server.query.get(self.target) or self.target)
            data.update(target=d)
            data.update(executor=str(self.executor) if self.executor else None)
            data.update(service=str(self.service) if self.service else None)
            if self.orchestration:
                data.update(
                    orchestration=dict(id=str(self.orchestration.id), name=self.orchestration.name,
                                       version=self.orchestration.version))
            else:
                data.update(
                    orchestration=None)
            if self.server:
                data.update(server=dict(id=str(self.server.id), name=self.server.name))
            else:
                data.update(server=None)
        else:
            data.update(target=self.target)
            data.update(
                orchestration_id=str(getattr(self.orchestration, 'id', None)) if getattr(self.orchestration, 'id',
                                                                                         None) else None)
            data.update(
                executor_id=str(getattr(self.executor, 'id', None)) if getattr(self.executor, 'id', None) else None)
            data.update(
                service_id=str(getattr(self.service, 'id', None)) if getattr(self.service, 'id', None) else None)
            data.update(
                server_id=str(getattr(self.server, 'id', None)) if getattr(self.server,
                                                                           'id', None) else None)
        data.update(params=self.params)
        data.update(success=self.success)
        data.update(undo_success=self.undo_success)
        data.update(message=self.message)

        if self.parent_step_execution_id and not add_step_exec:
            data.update(parent_step_execution_id=str(self.parent_step_execution_id))
        if add_step_exec:
            steps = []
            for se in self.step_executions:
                se: StepExecution

                se_json = se.to_json(human, split_lines=split_lines)
                if se.child_orch_execution:
                    se_json['orch_execution'] = se.child_orch_execution.to_json(add_step_exec=add_step_exec,
                                                                                split_lines=split_lines,
                                                                                human=human)
                elif se.child_orch_execution_id:
                    from dimensigon.web.network import get, Response
                    from dimensigon.network.auth import HTTPBearerAuth
                    from flask_jwt_extended import create_access_token
                    params = ['steps']
                    if human:
                        params.append('human')


                    try:
                        auth = HTTPBearerAuth(create_access_token(get_jwt_identity()))
                        resp = get(se.server, 'api_1_0.orchexecutionresource',
                                   view_data=dict(execution_id=se.child_orch_execution_id, params=params), auth=auth)
                    except Exception as e:
                        current_app.logger.exception(f"Exception while trying to acquire orch execution "
                                                     f"{se.child_orch_execution_id} from {se.server}")
                        resp = Response(exception=e)

                    if resp.ok:
                        se_json['orch_execution'] = resp.msg
                        se_json.pop('child_orch_execution_id', None)

                steps.append(se_json)
            # steps.sort(key=lambda x: x.start_time)
            data.update(steps=steps)
        return data

    @classmethod
    def from_json(cls, kwargs):
        if 'start_time' in kwargs:
            kwargs['start_time'] = datetime.strptime(kwargs.get('start_time'), defaults.DATETIME_FORMAT)
        if 'end_time' in kwargs:
            kwargs['end_time'] = datetime.strptime(kwargs.get('end_time'), defaults.DATETIME_FORMAT)
        try:
            o = cls.query.get(kwargs.get('id'))
        except RuntimeError as e:
            o = None
        if o:
            for k, v in kwargs.items():
                if getattr(o, k) != v:
                    setattr(o, k, v)
            return o
        else:
            return cls(**kwargs)




import datetime as dt
import typing as t

from flask import current_app
from flask_jwt_extended import create_access_token
from sqlalchemy import not_

from dimensigon import defaults
from dimensigon.domain.entities import Scope, Server
from dimensigon.network.auth import HTTPBearerAuth
from dimensigon.utils.helpers import get_now


def get_servers_from_scope(scope: Scope, bypass: t.Union[t.List[Server], Server] = None) -> t.List[Server]:
    """
    Returns the servers to lock for the related scope

    Parameters
    ----------
    scope: Scope

    Returns
    -------

    """
    quorum = []
    me = Server.get_current()
    if scope == scope.CATALOG:
        last_alive_ids = current_app.cluster_manager.cluster.get_delta_keepalive(
            dt.timedelta(minutes=defaults.REFRESH_PERIOD * defaults.COMA_NODE_FACTOR))
        if me.id not in last_alive_ids:
            last_alive_ids.append(me.id)
        all_query = Server.query.filter_by(l_ignore_on_lock=False)
        if isinstance(bypass, list):
            all_query = all_query.filter(not_(Server.id.in_([s.id for s in bypass])))
        elif isinstance(bypass, Server):
            all_query = all_query.filter_by(Server.id != bypass.id)
        n_servers = all_query.count()

        adult_query = all_query.filter(Server.created_on <= get_now() - defaults.ADULT_NODES)
        n_adult_nodes = adult_query.count()

        if n_adult_nodes == 0:
            if n_servers <= defaults.MIN_SERVERS_QUORUM:
                return all_query.all()
            else:
                return all_query.filter(Server.id.in_(last_alive_ids)).order_by(
                    Server.created_on).limit(defaults.MIN_SERVERS_QUORUM).all()

        elegible_query = adult_query.filter(Server.id.in_(last_alive_ids))
        n_elegible = elegible_query.count()

        if n_elegible <= defaults.MIN_SERVERS_QUORUM:
            return elegible_query.all()
        else:
            servers = elegible_query.order_by(Server.created_on).all()
            if len(servers) < defaults.MIN_SERVERS_QUORUM:
                return servers
            else:
                cost_dict = {}
                for server in servers:
                    if server.route and server.route.cost is not None:
                        if server.route.cost not in cost_dict:
                            cost_dict[server.route.cost] = []
                        cost_dict[server.route.cost].append(server)
                if len(cost_dict) > defaults.MIN_SERVERS_QUORUM:
                    for v in cost_dict.values():
                        v.sort(key=lambda x: (x.last_modified_at, x.name))
                        quorum.append(v[0])
                else:
                    quorum.extend(servers[0:defaults.MIN_SERVERS_QUORUM])
                me = Server.get_current()
                if me not in quorum:
                    quorum.append(me)
    elif scope == scope.UPGRADE:
        quorum = Server.get_current()
    return quorum


def get_root_auth():
    return HTTPBearerAuth(create_access_token('00000000-0000-0000-0000-000000000001'))
